import logging
from pasta_eln.UI.guiCommunicate import Communicate
from pasta_eln.UI.form import Form

def getTable(qtbot, comm, docType):
  table = None
  def receiveTable(t):
    nonlocal table
    table = t
  comm.backendThread.worker.beSendTable.connect(receiveTable)
  comm.uiRequestTable.emit(docType,'',True)
  while table is None:
    qtbot.wait(100)
  return table


def test_simple(qtbot, caplog):

  comm = Communicate('research')
  while comm.backendThread.worker.backend is None:
    qtbot.wait(100)

  table = getTable(qtbot, comm, 'measurement')
  docID = table[table['name']=='simple.png']['id'].values[0]
  dialog = Form(comm, {'id':docID})
  dialog.setMinimumSize(1024,800)
  dialog.show()
  qtbot.addWidget(dialog)
  while True:
    qtbot.wait(100)
    if len(dialog.doc) > 2 and dialog.docTypeComboBox.count()>2:
      break
  qtbot.wait(1000)
  path = qtbot.screenshot(dialog)
  print(path)
  comm.shutdownBackendThread()

  errors = [record for record in caplog.records if record.levelno >= logging.ERROR]
  assert not errors, f"Logging errors found: {[record.getMessage() for record in errors]}"



""" Sidebar widget that includes the navigation items """
import logging
from enum import Enum
from typing import Any
import pandas as pd
from anytree import Node
from PySide6.QtCore import Slot
from PySide6.QtGui import QResizeEvent
from PySide6.QtWidgets import QFrame, QLabel, QTreeWidgetItem, QVBoxLayout, QWidget
from ..backendWorker.worker import Task
from .guiCommunicate import Communicate
from .guiStyle import IconButton, TextButton, space, widgetAndLayout, widgetAndLayoutGrid


class Sidebar(QWidget):
  """ Sidebar widget that includes the navigation items """
  def __init__(self, comm:Communicate):
    super().__init__()
    self.comm = comm
    self.comm.changeSidebar.connect(self.paint)
    self.comm.backendThread.worker.beSendTable.connect(self.onGetData)
    self.projects = pd.DataFrame()
    self.sideBarWidth = self.comm.configuration['GUI']['sidebarWidth']

    # GUI elements
    mainL = QVBoxLayout()
    self.setFixedWidth(self.sideBarWidth)
    mainL.setContentsMargins(space['s'],space['s'],space['0'],space['s'])
    mainL.setSpacing(15)
    if self.comm.configuration['GUI']['showProjectBtn']=='Yes':
      TextButton('List projects', self, [Command.LIST_PROJECTS], mainL, 'Show list of all projects')
    _, self.projectsListL = widgetAndLayout('V', mainL, spacing='m')
    self.setLayout(mainL)

    self.widgetsAction:dict[str,QWidget] = {}
    self.widgetsList:dict[str,QWidget]   = {}
    self.widgetsProject:dict[str,Any]    = {}               #title bar and widget that contains all of project
    self.btnProjects:list[TextButton]    = []                               # list of buttons to show projects
    self.btnScan:TextButton|None         = None
    self.btnDocTypes:list[IconButton]    = []                         # list of buttons to show docType tables
    self.btnUnknown:IconButton|None      = None
    self.comm.uiRequestTable.emit('x0', '', True)


  @Slot(pd.DataFrame, str)
  def onGetData(self, projects:pd.DataFrame, docType:str) -> None:
    """
    Callback function to handle the received projects data

    Args:
      projects (pd.DataFrame): DataFrame containing project information
      docType (str): document type (should be 'x0' for projects for the sidebar)
    """
    if docType == 'x0':
      self.projects = projects
      self.paint('redraw')


  @Slot(str)
  def paint(self, projectChoice:str='') -> None:
    """
    Redraw sidebar: e.g. after change of project visibility in table

    Args:
      projectChoice (str): projectID on which to focus: '' string=draw default=none; 'redraw' implies redraw; id implies id
    """
    # Delete old widgets from layout and create storage
    for i in reversed(range(self.projectsListL.count())):
      self.projectsListL.itemAt(i).widget().setParent(None)
    if projectChoice != 'redraw':
      self.comm.projectID = projectChoice
    self.widgetsAction = {}
    self.widgetsList = {}
    self.widgetsProject = {}                                #title bar and widget that contains all of project
    # fill sidebar
    if self.projects.empty:
      self.projectsListL.addWidget(QLabel('Error: projects not received'))
      return
    if 'status' in self.projects.columns and len(self.projects)>5:
      temp = self.projects[self.projects['status']=='active']
      if len(temp)>2:
        self.projects = temp
    self.projects = self.projects.sort_values('name', axis=0).reset_index(drop=True)
    maxProjects = int((self.height()-120)/50)-1
    for index, project in self.projects.iterrows():
      if index>maxProjects:
        break
      projID = project['id']
      projName = project['name']
      #head: show project name as button
      projectW = QFrame()
      # projectW.setMinimumHeight(300) # convenience: allow scroll in sidebar
      projectL = QVBoxLayout(projectW)
      projectL.setSpacing(3)
      projectL.setContentsMargins(3,3,3,3)
      maxLabelCharacters = int((self.sideBarWidth-50)/7.1)
      label = (projName if len(projName) < maxLabelCharacters else f'{projName[:maxLabelCharacters - 3]}...')
      self.btnProjects.append(TextButton(label, self, [Command.SHOW_PROJECT, projID, ''], projectL))
      self.widgetsProject[projID] = [self.btnProjects[-1], projectW]

      # actions: scan, curate, ..
      actionW, actionL = widgetAndLayoutGrid(projectL)
      if self.comm.projectID != projID:                                       #depending which project is open
        actionW.hide()
        projectW.setStyleSheet(self.comm.palette.get('secondaryDark', 'background-color'))
      else:
        projectW.setStyleSheet(self.comm.palette.get('secondaryLight','background-color'))
      self.btnScan = TextButton('Scan', self, [Command.SCAN_PROJECT, projID], None, 'Scan', \
                            iconName='mdi.clipboard-search-outline')
      actionL.addWidget(self.btnScan, 0,0)
      # btnCurate = TextButton('Special', self, [projID], None)
      # btnCurate.hide()
      # actionL.addWidget(btnCurate, 0,1)
      self.widgetsAction[projID] = actionW

      # lists: view list of measurements, ... of this project
      listW, listL = widgetAndLayoutGrid(projectL,  spacing='s')
      if self.comm.projectID != projID:
        listW.hide()
      self.btnDocTypes = []
      for idx, (doctype, value) in enumerate(self.comm.docTypesTitles.items()):
        if doctype[0]!='x' and '/' not in doctype:
          icon = 'fa5s.asterisk' if value['icon']=='' else value['icon']
          self.btnDocTypes.append(IconButton(icon, self, [Command.LIST_DOCTYPE,doctype,projID], None, value['title']))
          listL.addWidget(self.btnDocTypes[-1], 0, idx)
      self.btnUnknown = IconButton('fa5.file', self, [Command.LIST_DOCTYPE,'-',projID], None, 'Unidentified')
      listL.addWidget(self.btnUnknown, 0, len(self.comm.docTypesTitles)+1)
      self.widgetsList[projID] = listW

      self.projectsListL.addWidget(projectW)
    # Other buttons
    stretch = QWidget()
    self.projectsListL.addWidget(stretch, stretch=2)
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks a button

    Args:
      command (list): list of commands
    """
    if command[0] is Command.LIST_DOCTYPE:
      self.comm.changeTable.emit(command[1], command[2])
    elif command[0] is Command.LIST_PROJECTS:
      self.comm.changeTable.emit('x0', '')
    elif command[0] is Command.SHOW_PROJECT:
      projID = command[1]
      item   = command[2]
      if item=='':                                                   #clicked on project-button, not tree view
        self.comm.projectID = projID
        for docID, widget in self.widgetsAction.items():
          if docID == projID:
            widget.show()
          else:
            widget.hide()
        for docID, widget in self.widgetsList.items():
          if docID == projID:
            widget.show()
          else:
            widget.hide()
        for docID, [_, projWidget] in self.widgetsProject.items():
          if docID == projID:
            projWidget.setStyleSheet(self.comm.palette.get('secondaryLight','background-color'))
          else:
            projWidget.setStyleSheet(self.comm.palette.get('secondaryDark','background-color'))
      self.comm.changeProject.emit(projID, item)
    elif command[0] is Command.SCAN_PROJECT:
      self.comm.uiRequestTask.emit(Task.SCAN, {'docID':self.comm.projectID})
    else:
      logging.error('Sidebar menu unknown: %s',command, exc_info=True)
    return


  def iterateTree(self, nodeHier:Node, projectID:str) -> QTreeWidgetItem:
    """
    Recursive function to translate the hierarchical node into a tree-node

    Args:
      nodeHier (Anytree.Node): anytree node
      projectID (str): project id of this tree

    Returns:
      QtTreeWidgetItem: tree node
    """
    #prefill with name, doctype, id
    nodeTree = QTreeWidgetItem([nodeHier.name, f'{projectID}/{nodeHier.id}'])
    children = []
    for childHier in nodeHier.children:
      if childHier.docType[0][0]=='x':
        childTree = self.iterateTree(childHier, projectID)
        children.append(childTree)
    if children:
      nodeTree.insertChildren(0,children)
    return nodeTree


  def resizeEvent(self, event: QResizeEvent) -> None:
    """
    executed upon resize

    Args:
      event (QResizeEvent): event
    """
    self.paint('redraw')
    return super().resizeEvent(event)


class Command(Enum):
  """ Commands used in this file """
  LIST_DOCTYPE = 1
  SHOW_PROJECT = 2
  SCAN_PROJECT = 3
  LIST_PROJECTS= 4



"""example add-on: plot data from files

THIS IS A VERY ADVANCED ADD-ON TUTORIAL
This tutorial teaches
- how to plot the data from files
"""
import matplotlib
import pandas as pd
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar
from PySide6.QtCore import QSize, Qt
from PySide6.QtWidgets import QComboBox, QDialog, QHBoxLayout, QLineEdit, QTableWidget, QTableWidgetItem
from pasta_eln.miscTools import MplCanvas, callDataExtractor, getDoc, isFloat
from pasta_eln.UI.guiStyle import Label, space, widgetAndLayout

# The following two variables are mandatory
description  = 'Default data plot'  #short description that is shown in the menu
reqParameter = {} #possibility for required parameters: like API-key, etc. {'API': 'text'}


class DataAnalyse(QDialog):
  """ Editor to change metadata of binary file """
  def __init__(self, df, data, widget):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
      df (pandas dataframe): all metadata
      data (list): all data
    """
    super().__init__()
    self.df = df
    self.data = data
        # GUI elements
    self.setWindowTitle('Analyze data')
    self.setMinimumWidth(1050)  #set size to match 4 blocks of 8bytes
    masterL = QHBoxLayout(self)
    masterL.setSpacing(space['s'])
    self.setStyleSheet(f"QLineEdit, QComboBox {{  {widget.comm.palette.get('secondaryText','color')} }}")
    _, mainL = widgetAndLayout('V', masterL, spacing='s')
    self.rightW = QTableWidget(self.df.shape[0], 1)
    for row in range(self.rightW.rowCount()):
      item = QTableWidgetItem(self.df['name'].iloc[row])
      item.setFlags(Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable)
      item.setCheckState(Qt.CheckState.Checked)
      self.rightW.setItem(row, 0, item)
    self.rightW.setStyleSheet('QTableView::indicator {width: 24px; height: 24px;}')
    self.rightW.verticalHeader().hide()
    self.rightW.horizontalHeader().hide()
    self.rightW.clicked.connect(self.cellClicked)
    self.rightW.setAlternatingRowColors(True)
    header = self.rightW.horizontalHeader()
    header.setStyleSheet('QHeaderView::section {padding: 0px 5px; margin: 0px;}')
    header.setStretchLastSection(True)
    masterL.addWidget(self.rightW)
    self.columns = list(data[0].columns)

    #graph
    self.graphW, graphL = widgetAndLayout('V', None)
    self.graph = MplCanvas(self, width=5, height=4, dpi=100)
    self.graphToolbar = NavigationToolbar(self.graph, self)
    self.graphToolbar.setIconSize(QSize(24, 24))
    self.graphToolbar.setStyleSheet('QToolButton {min-width: 18px; min-height: 24px;}')
    graphL.addWidget(self.graphToolbar)
    graphL.addWidget(self.graph)
    mainL.addWidget(self.graphW, stretch=1)

    #X-Axis
    _, rowXL = widgetAndLayout('H', mainL, 'm', 's', '0', 's')
    Label('x-axis:','h2',rowXL)
    self.xAxisCB = QComboBox()
    self.xAxisCB.addItems(self.columns)
    self.xAxisCB.currentTextChanged.connect(self.refresh)
    rowXL.addWidget(self.xAxisCB)
    Label('min:','',rowXL)
    self.xAxisMin = QLineEdit('')
    self.xAxisMin.textChanged.connect(self.refresh)
    rowXL.addWidget(self.xAxisMin)
    Label('max:','',rowXL)
    self.xAxisMax = QLineEdit('')
    self.xAxisMax.textChanged.connect(self.refresh)
    rowXL.addWidget(self.xAxisMax)
    Label('label:','',rowXL)
    self.xAxisLabel = QLineEdit('')
    self.xAxisLabel.textChanged.connect(self.refresh)
    rowXL.addWidget(self.xAxisLabel)

    #Y-Axis
    self.rowYW, rowYL = widgetAndLayout('H', mainL, 'm', 's', '0', 's')
    Label('y-axis:','h2',rowYL)
    self.yAxisCB = QComboBox()
    self.yAxisCB.addItems(self.columns)
    self.yAxisCB.currentTextChanged.connect(self.refresh)
    rowYL.addWidget(self.yAxisCB)
    Label('min:','',rowYL)
    self.yAxisMin = QLineEdit('')
    self.yAxisMin.textChanged.connect(self.refresh)
    rowYL.addWidget(self.yAxisMin)
    Label('max:','',rowYL)
    self.yAxisMax = QLineEdit('')
    self.yAxisMax.textChanged.connect(self.refresh)
    rowYL.addWidget(self.yAxisMax)
    Label('label:','',rowYL)
    self.yAxisLabel = QLineEdit('')
    self.yAxisLabel.textChanged.connect(self.refresh)
    rowYL.addWidget(self.yAxisLabel)

    self.refresh()

  def refresh(self):
    matplotlib.pyplot.close('all')
    self.graph.axes.cla()
    for idx, iData in enumerate(self.data):
      item = self.rightW.item(idx, 0)
      if item.checkState() != Qt.CheckState.Checked:
        continue
      x = iData[self.xAxisCB.currentText()]
      y = iData[self.yAxisCB.currentText()]
      label = list(self.df['name'])[idx]
      self.graph.axes.plot(x, y, label=label)
    self.graph.axes.legend()
    xMin = None if self.xAxisMin.text()=='' or not isFloat(self.xAxisMin.text()) else float(self.xAxisMin.text())
    xMax = None if self.xAxisMax.text()=='' or not isFloat(self.xAxisMin.text()) else float(self.xAxisMax.text())
    yMin = None if self.yAxisMin.text()=='' or not isFloat(self.xAxisMin.text()) else float(self.yAxisMin.text())
    yMax = None if self.yAxisMax.text()=='' or not isFloat(self.xAxisMin.text()) else float(self.yAxisMax.text())
    self.graph.axes.set_xlabel(self.xAxisLabel.text() or self.xAxisCB.currentText())
    self.graph.axes.set_ylabel(self.yAxisLabel.text() or self.yAxisCB.currentText())
    self.graph.axes.set_xlim(left=xMin, right=xMax)
    self.graph.axes.set_ylim(bottom=yMin, top=yMax)
    self.graph.axes.tick_params(axis='both', direction='in')
    self.graph.draw() # Trigger the canvas to update and redraw
    self.graphToolbar.show()
    self.graph.show()


  def cellClicked(self, item):
    """ after table is clicked: refresh graph """
    self.refresh()


def main(comm, df, widget, parameter={}):
    """ main function: has to exist and is called by the menu
    Args:
        comm (communication): pasta communication layer
        df (Dataframe): dataframe with the data to plot. The docIDand path columns are of most interest
        widget (QWidget): allows to create new gui dialogs
        parameter (dict): ability to pass parameters

    Returns:
        bool: success
    """
    # no cleaning needed since we only use id and path columns
    data = []
    for idx, row in df.iterrows():
        res = callDataExtractor(row['docID'], comm)
        if res is None:
            continue
        # Here you can process the extracted data further, e.g., plot it
        if isinstance(res, pd.DataFrame):
            data.append(res)
        else:
            print(f"Extracted data is not a DataFrame. {row['path']}")
    if data:
        dataAnalyse = DataAnalyse(df, data, widget)
        dataAnalyse.exec()
    return True



""" widget that shows the table of the items """
import itertools
import logging
import re
from enum import Enum
from pathlib import Path
from typing import Any
import pandas as pd
from PySide6.QtCore import QModelIndex, Qt, Slot
from PySide6.QtGui import QStandardItem, QStandardItemModel
from PySide6.QtWidgets import (QApplication, QComboBox, QFileDialog, QHeaderView, QMenu, QMessageBox, QTableView,
                               QVBoxLayout, QWidget)
from ..backendWorker.worker import Task
from ..miscTools import callAddOn
from .gallery import ImageGallery
from .guiCommunicate import Communicate
from .guiStyle import Action, Label, TextButton, space, widgetAndLayout, widgetAndLayoutGrid
from .tableFilterManager import FilterCommand, FilterManager
from .tableHeader import TableHeader


#Scan button to more button
class Table(QWidget):
  """ widget that shows the table of the items """
  def __init__(self, comm:Communicate):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
    """
    super().__init__()
    self.comm = comm
    self.comm.backendThread.worker.beSendTable.connect(self.onGetData)
    self.comm.changeTable.connect(self.changeTable)
    self.comm.stopSequentialEdit.connect(self.stopSequentialEditFunction)
    self.stopSequentialEdit = False
    self.data         :pd.DataFrame = pd.DataFrame()
    self.baseModel   :QStandardItemModel = QStandardItemModel()
    self.filterManager:FilterManager = FilterManager(self.baseModel)
    self.filterHeader:list[str] = []
    self.lastClickedRow = -1
    self.flagGallery = False
    self.docType = 'x0'
    self.showAll= self.comm.configuration['GUI']['showHidden']=='Yes'

    ### GUI elements
    mainL = QVBoxLayout()
    mainL.setSpacing(0)
    mainL.setContentsMargins(space['s'], space['s'], space['s'], space['s'])
    # header
    self.headerW, headerL = widgetAndLayout('H', mainL, 'm')
    self.headerW.hide()
    self.headline = Label('','h1', headerL)
    self.showState= Label('', 'h3', headerL)
    self.subDocTypeL= Label('      subType:', 'h3', headerL)
    self.subDocTypeL.hide()
    self.subDocType=QComboBox(self)
    self.subDocType.setMinimumWidth(200)
    self.subDocType.currentTextChanged.connect(lambda dt: self.changeTable(dt, self.comm.projectID))
    self.subDocType.hide()
    headerL.addWidget(self.subDocType)
    headerL.addStretch(1)
    self.addBtn = TextButton('Add',  self, [Command.ADD_ITEM],   headerL)

    self.selectionBtn = TextButton('Selection', self, [], headerL)
    selectionMenu = QMenu(self)
    Action('Toggle selection',self, [Command.TOGGLE_SELECTION], selectionMenu)
    selectionMenu.addSeparator()
    Action('Group Edit',      self, [Command.GROUP_EDIT],       selectionMenu)
    Action('Sequential edit', self, [Command.SEQUENTIAL_EDIT],  selectionMenu)
    self.toggleHidden  = Action('Invert hidden status of selected', self, [Command.TOGGLE_HIDE],    selectionMenu)
    Action('Rerun extractors',self, [Command.RERUN_EXTRACTORS], selectionMenu)
    Action('Delete',          self, [Command.DELETE],           selectionMenu)
    self.selectionBtn.setMenu(selectionMenu)

    self.visibilityBtn = TextButton('Visibility', self, [], headerL)
    visibilityMenu = QMenu(self)
    Action(                    'Add Filter',            self, [FilterCommand.ADD_FILTER],visibilityMenu)
    self.showHidden    = Action('Show/hide hidden ___', self, [Command.SHOW_ALL],        visibilityMenu)
    self.toggleGallery = Action('Toggle gallery view',  self, [Command.TOGGLE_GALLERY],  visibilityMenu)
    self.toggleGallery.setVisible(False)
    self.visibilityBtn.setMenu(visibilityMenu)

    self.btnMore = TextButton('More', self, [], headerL)
    self.moreMenu = QMenu(self)
    Action('Export to csv',            self, [Command.EXPORT],   self.moreMenu)
    self.moreMenu.addSeparator()
    projectGroup = self.comm.configuration['projectGroups'][self.comm.projectGroup]
    if projectAddOns := projectGroup.get('addOns',{}).get('table',''):
      for label, description in projectAddOns.items():
        Action(description, self, [Command.ADD_ON, label], self.moreMenu)
      self.moreMenu.addSeparator()
    self.actionChangeColums = Action('Change list columns',  self, [Command.CHANGE_COLUMNS], self.moreMenu)#add this action at end
    self.btnMore.setMenu(self.moreMenu)

    # filter
    _, self.filterL = widgetAndLayoutGrid(mainL, top='s', bottom='s')
    # table
    self.table = QTableView(self)
    self.table.setStyleSheet('QTableView::indicator {width: 16px; height: 16px;}')
    self.table.verticalHeader().hide()
    self.table.clicked.connect(self.cellClicked)
    self.table.doubleClicked.connect(self.cell2Clicked)
    self.table.setSortingEnabled(True)
    self.table.setAlternatingRowColors(True)
    header = self.table.horizontalHeader()
    header.setStyleSheet('QHeaderView::section {padding: 0px 5px; margin: 0px;}')
    header.setSectionsMovable(True)
    header.setSortIndicatorShown(True)
    header.setMaximumSectionSize(self.comm.configuration['GUI']['maxTableColumnWidth'])
    header.setStretchLastSection(True)
    mainL.addWidget(self.table)
    self.gallery = ImageGallery(self.comm)
    self.gallery.setVisible(False)
    mainL.addWidget(self.gallery)
    self.setLayout(mainL)
    self.setStyleSheet(f"QLineEdit, QComboBox {{ {self.comm.palette.get('secondaryText', 'color')} }}")
    self.comm.uiRequestTable.emit(self.docType, self.comm.projectID, self.showAll)


  @Slot(str, str)
  def changeTable(self, docType:str, projID:str) -> None:
    """ What happens when user clicks to change doc-type
    Args:
      docType (str): document type
      projID (str): project id
    """
    if docType:
      self.docType = docType
    if projID:
      self.comm.projectID  = projID
    if self.docType == 'x0':
      self.comm.projectID = ''
    logging.debug('request table for %s, %s %s', self.docType, self.comm.projectID, self.showAll)
    self.comm.uiRequestTable.emit(self.docType, self.comm.projectID, self.showAll)


  @Slot(pd.DataFrame, str)
  def onGetData(self, data:pd.DataFrame, docType:str) -> None:
    """
    Callback function to handle the received data

    Args:
      data (pd.DataFrame): DataFrame containing table
      docType (str): document type
    """
    logging.debug('got table data %s', docType)
    if docType == self.docType:
      self.data = data
      self.paint()


  @Slot()
  def paint(self) -> None:
    """
    What happens when the table changes its raw information
    """
    if self.isHidden():
      return

    # Clear existing filters and GUI elements
    self.filterManager.clearAll()
    for i in reversed(range(self.filterL.count())):
      item_1 = self.filterL.itemAt(i)
      if item_1 is not None:
        item_1.widget().setParent(None)
    allDocTypes:list[str] = []
    if '/' not in self.docType:
      # get list of all subDocTypes
      allDocTypes = [i for i in self.comm.docTypesTitles if i.startswith(self.docType)]
      if len(allDocTypes) <= 1:
        self.subDocTypeL.hide()
        self.subDocType.hide()
      else:
        self.subDocTypeL.show()
        self.subDocType.show()
        alreadyInside = {self.subDocType.itemText(i) for i in range(self.subDocType.count())}
        if alreadyInside != set(allDocTypes):
          self.subDocType.clear()
          self.subDocType.addItems(allDocTypes)
    if 'measurement' in self.docType:
      self.toggleGallery.setVisible(True)
    else:
      self.toggleGallery.setVisible(False)
      self.flagGallery = False
    # start tables: use data
    if self.docType=='_tags_':
      self.addBtn.hide()
      self.filterHeader = ['tag','name','type']
      self.headline.setText('TAGS')
      self.actionChangeColums.setVisible(False)
    else:
      self.addBtn.show()
      if self.docType.startswith('x0'):
        self.selectionBtn.hide()
        self.toggleHidden.setVisible(False)
      else:
        self.selectionBtn.show()
        self.toggleHidden.setVisible(True)
      self.showState.setText('(show all rows)' if self.showAll else '(hide hidden rows)')
      docLabel = 'Unidentified'
      if self.docType=='-':
        self.actionChangeColums.setVisible(False)
      else:
        self.actionChangeColums.setVisible(True)
        docLabel = self.comm.docTypesTitles[self.docType]['title']
      if self.comm.projectID:
        self.headline.setText(docLabel)
        self.showHidden.setText(f'Show/hide hidden {docLabel.lower()}')
      else:
        self.headline.setText(f'All {docLabel.lower()}')
        self.showHidden.setText(f'Show/hide all hidden {docLabel.lower()}')
      columnNames = [i.replace('metaUser.','u.') for i in self.data.columns]
      self.filterHeader = list(columnNames)[:-2]
    self.headerW.show()
    nRows, nCols = self.data.shape
    model = QStandardItemModel(nRows, nCols-2)
    model.setHorizontalHeaderLabels(self.filterHeader)
    for i, j in itertools.product(range(nRows), range(nCols-2)):
      value = self.data.iloc[i,j]
      if self.docType=='_tags_':                                                                   # tags list
        if j==0 and re.match(r'_\d', value):                                                            # star
          item = QStandardItem('\u2605'*int(value[1]))
        elif j==0:
          item = QStandardItem(value)
        else:
          item = QStandardItem(value)
      elif value in ('None','','nan'):                                                           # None, False
        item = QStandardItem('-')
      elif value=='True':                                                                               # True
        item = QStandardItem('Y')
      elif isinstance(value, str) and re.match(r'^[a-z\-]-[a-z0-9]{32}$',value):                        # Link
        item = QStandardItem('oo')
      else:
        if self.filterHeader[j]=='tags':
          tags = [i.strip() for i in value.split(',')]
          if elementStar:=list(filter(re.compile(r'^_\d$').match, tags)):
            tags.remove(elementStar[0])
            tags = ['\u2605'*int(elementStar[0][1])]+tags
          text = ' '.join(tags)
        else:
          text = value
        item = QStandardItem(text)
      if (j==0 and self.docType!='_tags_') or (self.docType=='_tags_' and j==1):
        if 'F' in self.data['show'][i]:
          item.setText(f'{item.text()}  \U0001F441')
        item.setAccessibleText(self.data['id'][i])
        if self.docType[0] != 'x':
          item.setFlags(Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable)
          item.setCheckState(Qt.CheckState.Unchecked)
      else:
        item.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable)
      model.setItem(i, j, item)

    # Update the filter manager with the new base model
    self.baseModel = model
    self.filterManager.setBaseModel(model)

    if self.flagGallery:
      self.gallery.updateGrid(model)
      self.gallery.setVisible(True)
      self.table.setVisible(False)
    else:
      self.table.setModel(self.filterManager.getFinalModel())
      self.table.horizontalHeader().resizeSections(QHeaderView.ResizeMode.ResizeToContents)
      self.table.horizontalHeader().setStretchLastSection(True)
      self.table.sortByColumn(0, Qt.SortOrder.AscendingOrder)
      self.table.show()
      self.gallery.setVisible(False)
      self.table.setVisible(True)
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.ADD_ITEM:
      self.comm.formDoc.emit({'type':[self.docType], '_projectID':self.comm.projectID})
      self.comm.changeTable.emit(self.docType, self.comm.projectID)
      if self.docType=='x0':
        self.comm.changeSidebar.emit('redraw')

    elif command[0] is Command.GROUP_EDIT:
      docIDs = []
      finalModel = self.filterManager.getFinalModel()
      for row in range(finalModel.rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          docIDs.append(docID)
      if docIDs:
        self.comm.formDoc.emit({'type':[self.docType], '_ids':docIDs})
        self.changeTable(self.docType, self.comm.projectID)

    elif command[0] is Command.SEQUENTIAL_EDIT:
      self.stopSequentialEdit = False
      finalModel = self.filterManager.getFinalModel()
      for row in range(finalModel.rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          self.comm.formDoc.emit({'id':docID})
        if self.stopSequentialEdit:
          break
      self.comm.changeTable.emit(self.docType, self.comm.projectID)

    elif command[0] is Command.DELETE:
      ret = None
      finalModel = self.filterManager.getFinalModel()
      for row in range(finalModel.rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          if ret is None:
            ret = QMessageBox.critical(self, 'Warning', 'Are you sure you want to delete the data?',
                QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No)
          if ret==QMessageBox.StandardButton.Yes:
            self.comm.uiRequestTask.emit(Task.DELETE_DOC, {'docID':docID})
      self.comm.changeTable.emit(self.docType, self.comm.projectID)

    elif command[0] is Command.CHANGE_COLUMNS:
      dialog = TableHeader(self.comm, self.docType)
      dialog.exec()
      self.comm.uiRequestTable.emit(self.docType, self.comm.projectID, self.showAll)

    elif command[0] is Command.EXPORT:
      fileName = QFileDialog.getSaveFileName(self,'Export to ..',str(Path.home()),'*.csv')[0]
      if not fileName.endswith('.csv'):
        fileName += '.csv'
      finalModel = self.filterManager.getFinalModel()
      with open(fileName,'w', encoding='utf-8') as fOut:
        header = [f'"{i}"' for i in self.filterHeader]
        fOut.write(','.join(header)+'\n')
        for row in range(finalModel.rowCount()):
          rowContent = []
          for col in range(finalModel.columnCount()):
            value = finalModel.index( row, col, QModelIndex()).data(Qt.ItemDataRole.DisplayRole)
            rowContent.append(f'"{value}"')
          fOut.write(','.join(rowContent)+'\n')

    elif command[0] is Command.ADD_ON:
      # check if one is selected, if yes, only export selected; otherwise use All
      finalModel = self.filterManager.getFinalModel()
      useAll = True
      for row in range(finalModel.rowCount()):
        item, _ = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          useAll = False
          break
      data   = []
      for row in range(finalModel.rowCount()):
        item, docID = self.itemFromRow(row)
        if useAll or  item.checkState() == Qt.CheckState.Checked:
          dataRow = [docID]
          for col in range(finalModel.columnCount()):
            value = finalModel.index(row, col).data(Qt.ItemDataRole.DisplayRole)
            dataRow.append(value)
          data.append(dataRow)
      df = pd.DataFrame(data, columns=['docID']+self.filterHeader)
      callAddOn(command[1], self.comm, df, self)

    elif command[0] is Command.TOGGLE_HIDE:
      changeFlag = False
      finalModel = self.filterManager.getFinalModel()
      for row in range(finalModel.rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          self.comm.uiRequestTask.emit(Task.HIDE_SHOW, {'docID':docID})
          changeFlag = True
      if changeFlag:
        self.comm.uiRequestTable.emit(self.docType, self.comm.projectID, self.showAll)
      if self.docType=='x0':
        self.comm.changeSidebar.emit('redraw')
      self.paint()

    elif command[0] is Command.TOGGLE_SELECTION:
      finalModel = self.filterManager.getFinalModel()
      for row in range(finalModel.rowCount()):
        item,_ = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          item.setCheckState(Qt.CheckState.Unchecked)
        else:
          item.setCheckState(Qt.CheckState.Checked)

    elif command[0] is Command.SHOW_ALL:
      self.showAll = not self.showAll
      self.comm.uiRequestTable.emit(self.docType, self.comm.projectID, self.showAll)

    elif command[0] is Command.RERUN_EXTRACTORS:
      docIDs = []
      finalModel = self.filterManager.getFinalModel()
      for row in range(finalModel.rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          docIDs.append(docID)
      self.comm.uiRequestTask.emit(Task.EXTRACTOR_RERUN, {'docIDs':docIDs,'recipe':''})
      self.comm.uiRequestTable.emit(self.docType, self.comm.projectID, self.showAll)

    elif command[0] is Command.TOGGLE_GALLERY:
      self.flagGallery = not self.flagGallery
      self.paint()

    elif command[0] is FilterCommand.ADD_FILTER:
      # Create new filter using FilterManager
      filterItem = self.filterManager.addFilter(self.filterHeader, self)

      # Create GUI row
      _, rowL = widgetAndLayout('H', self.filterL, 'm', 'xl', '0', 'xl')
      rowL.addWidget(filterItem.select)
      rowL.addWidget(filterItem.text)
      rowL.addWidget(filterItem.inverse)
      rowL.addWidget(filterItem.delete)

      # Connect signals
      filterItem.select.currentIndexChanged.connect(lambda idx: self.filterChoice(idx, filterItem.id))
      filterItem.text.textChanged.connect(lambda text: self.filterTextChanged(text, filterItem.id))

      # Update table model
      self.table.setModel(self.filterManager.getFinalModel())

    elif command[0] is FilterCommand.DELETE_FILTER:
      filterID = command[1]

      # Find and remove GUI elements for this filter
      for i in range(self.filterL.count()):
        itemI = self.filterL.itemAt(i)
        if itemI and itemI.widget():
          widget = itemI.widget()
          layout = widget.layout()
          if layout and layout.count() >= 4:
            # Check if this is the right filter by looking at delete button's command
            itemJ = layout.itemAt(3)
            if itemJ is not None:
              delete_btn = itemJ.widget()
              if hasattr(delete_btn, 'command') and delete_btn.command[1] == filterID:
                widget.setParent(None)
                break

      # Remove filter from manager
      self.filterManager.removeFilter(filterID)

      # Update table model
      self.table.setModel(self.filterManager.getFinalModel())

    elif command[0] is FilterCommand.SET_FILTER:
      filterID = command[1]
      if len(command) > 2 and command[2] == 'invert':
        if filterItemI := self.filterManager.getFilter(filterID):
          self.filterTextChanged(filterItemI.text.text(), filterID)

    else:
      logging.error('Menu unknown: %s',command, exc_info=True)
    self.setStyleSheet(f"QLineEdit, QComboBox {{ {self.comm.palette.get('secondaryText', 'color')} }}")
    return


  def cellClicked(self, index: QModelIndex) -> None:
    """
    What happens when user clicks cell in table of tags, projects, samples, ..
    -> show details

    Args:
      index (QModelIndex): cell clicked
    """
    row = index.row()
    _, docID = self.itemFromRow(row)

    # Check if shift is held and lastClickedRow is set
    modifiers = QApplication.keyboardModifiers()
    if modifiers == Qt.ShiftModifier and self.lastClickedRow > -1:                # type: ignore[attr-defined]
      start = min(self.lastClickedRow, row)
      end = max(self.lastClickedRow, row)
      target_state = Qt.CheckState.Checked if self.itemFromRow(row)[0].checkState() == Qt.CheckState.Checked \
                      else Qt.CheckState.Unchecked
      for r in range(start, end + 1):
        item, _ = self.itemFromRow(r)
        item.setCheckState(target_state)
    else:                                             # No need to toggle only the clicked row, just record it
      self.lastClickedRow = row
    # Change view
    if docID[0] == 'x':                                                      # only show items for non-folders
      if self.docType == 'x0':
        self.comm.changeProject.emit(docID, '')
        self.comm.changeSidebar.emit(docID)
      else:
        self.comm.changeProject.emit(self.comm.projectID, docID)
        self.comm.changeSidebar.emit(self.comm.projectID)
    else:
      self.comm.changeDetails.emit(docID)
    return


  def cell2Clicked(self, index: QModelIndex) -> None:
    """
    What happens when user double clicks cell in table of projects

    Args:
      index (QModelIndex): cell clicked
    """
    row = index.row()
    _, docID = self.itemFromRow(row)
    if self.docType=='x0':
      self.comm.changeProject.emit(docID, '')
      self.comm.changeSidebar.emit(docID)
    else:
      self.comm.formDoc.emit({'id': docID})
      self.comm.changeTable.emit('','')
      self.comm.changeDetails.emit('redraw')
    return


  @Slot()
  def stopSequentialEditFunction(self) -> None:
    """ Stop the sequential edit of a number of items """
    self.stopSequentialEdit=True
    return


  def itemFromRow(self, row:int) -> tuple[QStandardItem, str]:
    """
    get item from row by iterating through the proxyModels

    Args:
      row (int): row number

    Returns:
      QItem, str: the item and docID
    """
    finalModel = self.filterManager.getFinalModel()
    index = finalModel.index(row, 0)
    # Map through all proxy models back to base model
    currentModel = finalModel
    while hasattr(currentModel, 'mapToSource') and currentModel != self.baseModel:
      index = currentModel.mapToSource(index)
      currentModel = currentModel.sourceModel()                                                 # type: ignore
    item = self.baseModel.itemFromIndex(index)
    return item, item.accessibleText()


  def filterChoice(self, item:int, filterID:int) -> None:
    """
    Change the column which is used for filtering

    Args:
       item (int): column number to filter by
       filterID (int): ID of the filter being modified
    """
    filterItem = self.filterManager.getFilter(filterID)
    if not filterItem:
      return
    if item == len(self.filterHeader):                                                               # ratings
      item = self.filterHeader.index('tag')
    filterItem.model.setFilterKeyColumn(item)
    filterItem.text.setText('')
    return


  def filterTextChanged(self, text:str, filterID:int) -> None:
    """ text in line-edit in the filter is changed: update regex

    Args:
      text (str): filter text
      filterID (int): ID of the filter being modified
    """
    filterItem = self.filterManager.getFilter(filterID)
    if not filterItem:
      return

    regexStr = text
    if '*' in regexStr:
      regexStr = regexStr.replace('*','\u2605')
      if filterItem.inverse.isChecked():
        regexStr = f'((?!((?<!\u2605){regexStr}(?!\u2605))).)*'
      else:
        regexStr = f'(?<!\u2605){regexStr}(?!\u2605)'
    elif filterItem.inverse.isChecked():
      regexStr = f'^((?!{regexStr}).)*$'
    filterItem.model.setFilterRegularExpression(regexStr)
    return



class Command(Enum):
  """ Commands used in this file """
  ADD_ITEM         = 2
  TOGGLE_SELECTION = 3
  GROUP_EDIT       = 4
  SEQUENTIAL_EDIT  = 5
  TOGGLE_HIDE      = 6
  RERUN_EXTRACTORS = 7
  DELETE           = 8
  SHOW_ALL         = 9
  EXPORT           = 10
  CHANGE_COLUMNS   = 11
  TOGGLE_GALLERY   = 14
  ADD_ON           = 15



import logging
from pasta_eln.UI.guiCommunicate import Communicate
from pasta_eln.UI.details import Details
from .test_34_GUI_Form import getTable

def test_simple(qtbot, caplog):

  comm = Communicate('research')
  while comm.backendThread.worker.backend is None:
    qtbot.wait(100)
  window = Details(comm)
  window.setMinimumSize(300, 800)
  window.show()
  qtbot.addWidget(window)

  table = getTable(qtbot, comm, 'measurement')
  docIDs = table['id'].values[:3].tolist()
  print(docIDs)

  for i in range(3):
    window.comm.changeDetails.emit(docIDs[i])
    qtbot.wait(1000)
    path = qtbot.screenshot(window)
    print(path)

  comm.shutdownBackendThread()

  errors = [record for record in caplog.records if record.levelno >= logging.ERROR]
  assert not errors, f"Logging errors found: {[record.getMessage() for record in errors]}"



""" All hash tools for pasta-eln backend worker """
import logging
import os
from io import BufferedReader
from pathlib import Path
from urllib import request


def generic_hash(path:Path, forceFile:bool=False) -> str:
  """
  Hash an object based on its mode

  inspired by:
  https://github.com/chris3torek/scripts/blob/master/githash.py

  Example implementation:
      result = generic_hash(sys.argv[1])
      print('%s: hash = %s' % (sys.argv[1], result))

  Args:
    path (Path): path
    forceFile (bool): force to get shasum of file and not of link (False for gitshasum)

  Returns:
    string: shasum

  Raises:
    ValueError: shasum of directory not supported
  """
  if str(path).startswith('http'):                                                               #Remote file:
    try:
      req = request.Request(path.as_posix().replace(':/','://'), headers={'User-Agent': 'Mozilla/5.0'})
      with request.urlopen(req, timeout=60) as site:
        meta = site.headers
        size = int(meta.get_all('Content-Length')[0])
        return blob_hash(site, size)
    except Exception:
      logging.error('Could not download content / hashing issue %s',path.as_posix().replace(':/','://'), exc_info=True)
      return ''
  if path.is_dir():
    raise ValueError(f'This seems to be a directory {path.as_posix()}')
  if forceFile and path.is_symlink():
    path = path.resolve()
  shasum = ''
  if path.is_symlink():                                                                #if link, hash the link
    shasum = symlink_hash(path)
  elif path.is_file():                                                                             #Local file
    with open(path, 'rb') as stream:
      shasum = blob_hash(stream, path.stat().st_size)
  return shasum


def symlink_hash(path:Path) -> str:
  """
  Return (as hash instance) the hash of a symlink
  Caller must use hexdigest() or digest() as needed on
  the result

  Args:
    path (string): path to symlink

  Returns:
    string: shasum of link, aka short string
  """
  from hashlib import sha1
  hasher = sha1()
  data = os.readlink(path).encode('utf8', 'surrogateescape')
  hasher.update(b'blob {len(data)}\0')
  hasher.update(data)
  return hasher.hexdigest()


def blob_hash(stream:BufferedReader, size:int) -> str:
  """
  Return (as hash instance) the hash of a blob,
  as read from the given stream

  Args:
    stream (string): content to be hashed
    size (int): size of the content

  Returns:
    string: shasum

  Raises:
    ValueError: size given is not the size of the stream
  """
  from hashlib import sha1
  hasher = sha1()
  hasher.update(f'blob {size}\0'.encode('ascii'))
  nRead = 0
  while True:
    data = stream.read(65536)                                    # read 64K at a time for storage requirements
    if data == b'':
      break
    nRead += len(data)
    hasher.update(data)
  if nRead != size:
    raise ValueError(f'{stream.name}: expected {size} bytes, found {nRead} bytes')
  return hasher.hexdigest()



import logging
from pasta_eln.UI.guiCommunicate import Communicate
from pasta_eln.UI.project import Project
from .test_34_GUI_Form import getTable

def test_simple(qtbot, caplog):

  comm = Communicate('research')
  while comm.backendThread.worker.backend is None:
    qtbot.wait(100)
  window = Project(comm)
  window.setMinimumSize(1024,800)
  window.show()

  df = getTable(qtbot, comm, 'x0')
  projID = df[df['name']=='PASTAs Example Project']['id'].values[0]
  comm.changeProject.emit(projID, '')
  qtbot.addWidget(window)
  qtbot.wait(1000)

  path = qtbot.screenshot(window)
  print(path)
  comm.shutdownBackendThread()

  errors = [record for record in caplog.records if record.levelno >= logging.ERROR]
  assert not errors, f"Logging errors found: {[record.getMessage() for record in errors]}"



""" Common functions in a number of widgets """
import logging
import os
import platform
import subprocess
from enum import Enum
from pathlib import Path
from typing import Any
from PySide6.QtCore import QPoint
from PySide6.QtWidgets import QMenu, QWidget
from ..backendWorker.worker import Task
from .guiStyle import Action


def initContextMenu(widget:QWidget, pos:QPoint) -> None:
  """
  Create a context menu

  Args:
    widget (QWidget): parent widget
    pos (position): Position to create context menu at
  """
  context = QMenu(widget)
  # for extractors
  projectGroup = widget.comm.projectGroup                                         # type: ignore[attr-defined]
  extractors = widget.comm.configuration['projectGroups'][projectGroup]['addOns']['extractors']# type: ignore[attr-defined]
  extension = '' if widget.doc['branch'][0]['path'] is None else Path(widget.doc['branch'][0]['path']).suffix[1:]# type: ignore[attr-defined]
  if extension.lower() in extractors:
    extractors = extractors[extension.lower()]
    baseDocType= widget.doc['type'][0]                                            # type: ignore[attr-defined]
    choices= {key:value for key,value in extractors.items() if key.startswith(baseDocType)}
    for key,value in choices.items():
      Action(value,                     widget, [CommandMenu.CHANGE_EXTRACTOR, key], context)
    context.addSeparator()
    Action('Save image',                widget, [CommandMenu.SAVE_IMAGE],            context)
  # Action('Open file with another application', widget.changeExtractor, context, widget, name='_openExternal_')
  Action('Open folder in file browser', widget, [CommandMenu.OPEN_FILEBROWSER],      context)
  Action('Hide',                        widget, [CommandMenu.HIDE],                  context)
  context.exec(widget.mapToGlobal(pos))
  return


def executeContextMenu(widget:QWidget, command:list[Any]) -> bool:
  """
  Execute context menu command

  Args:
    widget (QWidget): parent widget
    command (list): command

  Returns:
    bool: success
  """
  filePath = Path(widget.doc['branch'][0]['path'])                                # type: ignore[attr-defined]
  if command[0] is CommandMenu.OPEN_FILEBROWSER or command[0] is CommandMenu.OPEN_EXTERNAL:
    filePath = widget.comm.basePath/filePath                                      # type: ignore[attr-defined]
    filePath = filePath if command[0] is CommandMenu.OPEN_EXTERNAL else filePath.parent
    if platform.system() == 'Darwin':                                                                  # macOS
      subprocess.call(('open', filePath))
    elif platform.system() == 'Windows':                                                             # Windows
      os.startfile(filePath)                                                      # type: ignore[attr-defined]
    else:                                                                                     # linux variants
      subprocess.call(('xdg-open', filePath))
  elif command[0] is CommandMenu.SAVE_IMAGE:
    image = widget.doc['image']                                                   # type: ignore[attr-defined]
    if image.startswith('data:image/'):
      imageType = image[11:14] if image[14]==';' else image[11:15]
    else:
      imageType = 'svg'
    saveFilePath = widget.comm.basePath/filePath.parent/f'{filePath.stem}_PastaExport.{imageType.lower()}'# type: ignore[attr-defined]
    path = widget.doc['branch'][0]['path']                                        # type: ignore[attr-defined]
    if not path.startswith('http'):
      path = (widget.comm.basePath/path).as_posix()                               # type: ignore[attr-defined]
    widget.comm.uiRequestTask.emit(Task.EXTRACTOR_TEST,{'fileName':path,'recipe':'/'.join(widget.doc['type']),# type: ignore[attr-defined]
                                                        'saveFig':str(saveFilePath), 'style':''})
  elif command[0] is CommandMenu.HIDE:
    widget.comm.uiRequestTask.emit(Task.HIDE_SHOW, {'docID':widget.docID})        # type: ignore[attr-defined]
  elif command[0] is CommandMenu.CHANGE_EXTRACTOR:
    widget.comm.uiRequestTask.emit(Task.EXTRACTOR_RERUN, {'docIDs':[widget.doc['id']],'recipe':command[1]})# type: ignore[attr-defined]
  else:
    logging.error('Command not found in _contextMenu %s', command, exc_info=True)
  return True


class CommandMenu(Enum):
  """ Commands used in this file """
  CHANGE_EXTRACTOR = 2
  SAVE_IMAGE       = 3
  OPEN_FILEBROWSER = 4
  OPEN_EXTERNAL    = 5
  HIDE             = 6



""" Graphical user interface houses all widgets """
import json
import logging
import re
import sys
import webbrowser
from enum import Enum
from pathlib import Path
from typing import Any
from PySide6.QtCore import QEvent, QUrl, Slot
from PySide6.QtGui import QDesktopServices, QIcon, QPixmap, QShortcut
from PySide6.QtWidgets import QFileDialog, QLabel, QMainWindow
from pasta_eln import __version__
from ..backendWorker.worker import Task
from ..fixedStringsJson import CONF_FILE_NAME, AboutMessage, shortcuts
from ..miscTools import hardRestart, installPythonPackages, updateAddOnList
from .body import Body
from .config.main import Configuration
from .data_hierarchy.editor import SchemeEditor
from .definitions.editor import Editor as DefinitionsEditor
from .form import Form
from .guiCommunicate import Communicate
from .guiStyle import Action, ScrollMessageBox, widgetAndLayout
from .messageDialog import showMessage
from .palette import Palette
from .repositories.uploadGUI import UploadGUI
from .sidebar import Sidebar


class MainWindow(QMainWindow):
  """ Graphical user interface includes all widgets """

  def __init__(self, comm:Communicate) -> None:
    """ Init main window
    Args:
      projectGroup (str): project group to load
    """
    # global setting
    super().__init__()
    self.comm = comm
    if self.comm.configuration:
      self.comm.palette = Palette(self, self.comm.configuration['GUI']['theme'])
      self.comm.docTypesChanged.connect(self.paint)
    else:
      configWindow = Configuration(self.comm, 'setup')
      configWindow.exec()
      self.setCentralWidget(QLabel('ERROR: No configuration present!'))
      return
    self.comm.formDoc.connect(self.formDoc)
    self.comm.changeSidebar.connect(self.paint)
    self.comm.backendThread.worker.beSendTaskReport.connect(self.showReport)

    # GUI
    self.setWindowTitle(f"PASTA-ELN {__version__}")
    self.resize(self.screen().size())                                 #self.setWindowState(Qt.WindowMaximized)
    #TODO https://bugreports.qt.io/browse/PYSIDE-2706 https://bugreports.qt.io/browse/QTBUG-124892
    resourcesDir = Path(__file__).parent / 'Resources'
    self.setWindowIcon(QIcon(QPixmap(resourcesDir / 'Icons' / 'favicon64.png')))
    menu = self.menuBar()
    projectMenu = menu.addMenu('&Project')
    Action('&Export project to .eln',        self, [Command.EXPORT],         projectMenu)
    Action('&Import .eln into project',      self, [Command.IMPORT],         projectMenu)
    Action('&Upload to repository',          self, [Command.REPOSITORY],     projectMenu)
    Action('&Exit',                          self, [Command.EXIT],           projectMenu)

    self.viewMenu = menu.addMenu('Common &Lists')

    systemMenu = menu.addMenu('Project &group')
    self.changeProjectGroups = systemMenu.addMenu('&Change project group')
    syncMenu = systemMenu.addMenu('&Synchronize')
    Action('Send',                         self, [Command.SYNC_SEND],        syncMenu, shortcut='F5')
    if 'develop' in self.comm.configuration:
      Action('Get',                          self, [Command.SYNC_GET],       syncMenu, shortcut='F4')
      # Action('Smart synce',                  self, [Command.SYNC_SMART],       syncMenu)
    Action('&Item type editor',              self, [Command.SCHEMA],         systemMenu, shortcut='F8')
    Action('&Definitions editor',            self, [Command.DEFINITIONS],    systemMenu)
    systemMenu.addSeparator()
    Action('&Test extraction from a file',   self, [Command.TEST1],          systemMenu)
    Action('Test &selected item extraction', self, [Command.TEST2],          systemMenu, shortcut='F2')
    Action('Update &Add-on list',            self, [Command.UPDATE],         systemMenu)
    if 'develop' in self.comm.configuration:
      systemMenu.addSeparator()
      Action('&Verify database',             self, [Command.CHECK_DB],       systemMenu, shortcut='Ctrl+?')

    helpMenu = menu.addMenu('&Help')
    Action('&Website',                       self, [Command.WEBSITE],        helpMenu)
    Action('Shortcuts',                      self, [Command.SHORTCUTS],      helpMenu)
    Action('About',                          self, [Command.ABOUT],          helpMenu)
    systemMenu.addSeparator()
    Action('&Configuration',                 self, [Command.CONFIG],         helpMenu, shortcut='Ctrl+0')

    # shortcuts for advanced usage (user should not need)
    QShortcut('F9', self, lambda: self.execute([Command.RESTART]))
    if 'develop' not in self.comm.configuration:
      QShortcut('Ctrl+?', self, lambda: self.execute([Command.CHECK_DB]))

    # GUI elements
    mainWidget, mainLayout = widgetAndLayout('H')
    self.setCentralWidget(mainWidget)                                   # Set the central widget of the Window
    body = Body(self.comm)                                                             # body with information
    self.sidebar = Sidebar(self.comm)                                                   # sidebar with buttons
    mainLayout.addWidget(self.sidebar)
    mainLayout.addWidget(body)
    self.paint()


  @Slot(str)
  def paint(self, _:str='') -> None:
    """ Process things that might change """
    # Things that are inside the List menu
    self.viewMenu.clear()
    for key, value in self.comm.docTypesTitles.items():
      shortcut = None if value['shortcut']=='' else f"Ctrl+{value['shortcut']}"
      Action(value['title'],            self, [Command.VIEW, key],  self.viewMenu, shortcut=shortcut)
    self.viewMenu.addSeparator()
    Action('&Tags',               self, [Command.VIEW, '_tags_'], self.viewMenu, shortcut='Ctrl+T')
    Action('&Unidentified',       self, [Command.VIEW, '-'],      self.viewMenu, shortcut='Ctrl+U')
    # Things that are related to project group
    self.changeProjectGroups.clear()
    for name in self.comm.configuration['projectGroups'].keys():
      Action(name,                         self, [Command.CHANGE_PG, name], self.changeProjectGroups)
    return


  def closeEvent(self, event:QEvent) -> None:
    """
    Handle window close event - cleanup of backend thread

    Args:
      event: close event
    """
    if self.comm and hasattr(self.comm, 'backendThread') and self.comm.backendThread:
      self.comm.shutdownBackendThread()
    event.accept()


  @Slot(dict)
  def formDoc(self, doc: dict[str, Any]) -> None:
    """
    What happens when new/edit dialog is shown

    Args:
      doc (dict): document
    """
    formWindow = Form(self.comm, doc)
    ret = formWindow.exec()
    if ret == 0:
      self.comm.stopSequentialEdit.emit()
    return


  def execute(self, command: list[Any]) -> None:
    """
    action after clicking menu item
    """
    # file menu
    if command[0] is Command.EXPORT:
      if self.comm.projectID == '':
        showMessage(self, 'Error', 'You have to open a project to export', 'Critical')
        return
      fileName = QFileDialog.getSaveFileName(self, 'Save project into .eln file', str(Path.home()), '*.eln')[0]
      if fileName != '':
        docTypes = [i for i in self.comm.docTypesTitles if i[0]!='x']
        self.comm.uiRequestTask.emit(Task.EXPORT_ELN, {'fileName':fileName, 'projID':self.comm.projectID, 'docTypes':docTypes})
    elif command[0] is Command.IMPORT:
      if self.comm.projectID == '':
        showMessage(self, 'Error', 'You have to open a project to import', 'Critical')
        return
      fileName = QFileDialog.getOpenFileName(self, 'Load data from .eln file', str(Path.home()), '*.eln')[0]
      if fileName != '':
        self.comm.uiRequestTask.emit(Task.IMPORT_ELN, {'fileName':fileName, 'projID':self.comm.projectID})
        self.comm.changeProject.emit(self.comm.projectID, '')
    elif command[0] is Command.REPOSITORY:
      if self.comm.projectID == '':
        showMessage(self, 'Error', 'You have to open a project to upload', 'Critical')
        return
      dialogR = UploadGUI(self.comm)
      dialogR.exec()
    elif command[0] is Command.EXIT:
      self.close()
    # view menu
    elif command[0] is Command.VIEW:
      self.comm.projectID = ''
      self.comm.changeTable.emit(command[1], '')
      self.comm.changeSidebar.emit('')
    # system menu
    elif command[0] is Command.CHANGE_PG:
      self.comm.configuration['defaultProjectGroup'] = command[1]
      with open(Path.home()/CONF_FILE_NAME, 'w', encoding='utf-8') as fConf:
        fConf.write(json.dumps(self.comm.configuration, indent=2))
      self.comm.start(command[1])
    elif command[0] is Command.SYNC_SEND:
      self.comm.uiRequestTask.emit(Task.SEND_ELAB,  {'projGroup':self.comm.projectGroup})
    elif command[0] is Command.SYNC_GET:
      self.comm.uiRequestTask.emit(Task.GET_ELAB,   {'projGroup':self.comm.projectGroup})
    elif command[0] is Command.SYNC_SMART:
      self.comm.uiRequestTask.emit(Task.SMART_ELAB, {'projGroup':self.comm.projectGroup})
    elif command[0] is Command.SCHEMA:
      dialogS = SchemeEditor(self.comm)
      dialogS.exec()
    elif command[0] is Command.DEFINITIONS:
      dialogD = DefinitionsEditor(self.comm)
      dialogD.show()
    elif command[0] is Command.TEST1:
      fileName = QFileDialog.getOpenFileName(self, 'Open file for extractor test', str(Path.home()), '*.*')[0]
      if fileName is not None:
        self.comm.uiRequestTask.emit(Task.EXTRACTOR_TEST, {'fileName':fileName, 'style':'html', 'recipe':'', 'saveFig':''})
    elif command[0] is Command.TEST2:
      self.comm.testExtractor.emit()
    elif command[0] is Command.UPDATE:
      configProjecGroup = self.comm.configuration['projectGroups'][self.comm.projectGroup]
      installPythonPackages(configProjecGroup['addOnDir'])
      reportDict = updateAddOnList(self.comm.projectGroup)
      messageWindow = ScrollMessageBox('Add-on list updated', reportDict,
                                       style='QScrollArea{min-width:600 px; min-height:400px}')
      messageWindow.exec()
      hardRestart()
    elif command[0] is Command.CONFIG:
      dialogC = Configuration(self.comm)
      dialogC.exec()
    # remainder
    elif command[0] is Command.WEBSITE:
      webbrowser.open('https://pasta-eln.github.io/pasta-eln/')
    elif command[0] is Command.CHECK_DB:
      self.comm.uiRequestTask.emit(Task.CHECK_DB, {'style':'html'})
    elif command[0] is Command.SHORTCUTS:
      showMessage(self, 'Keyboard shortcuts', shortcuts, 'Information')
    elif command[0] is Command.ABOUT:
      showMessage(self, 'About', f'{AboutMessage}Environment: {sys.prefix}\n','Information')
    elif command[0] is Command.RESTART:
      hardRestart()
    else:
      logging.error('Gui menu unknown: %s', command, exc_info=True)
    return


  @Slot(Task, str, str, str)
  def showReport(self, task:Task, reportText:str, image:str, path:str) -> None:
    """ Show a report from backend worker
    Args:
      task (Task): task name
      reportText (str): text of the report
      image (str): base64 encoded image, svg image
      path (str): path to the file/folder that should be opened
    """
    if task is Task.OPEN_EXTERNAL and path:
      QDesktopServices.openUrl(QUrl.fromLocalFile(path))
      return
    if task in (Task.SCAN, Task.DROP_EXTERNAL):
      self.comm.changeProject.emit(self.comm.projectID, '')
    elif task is Task.CHECK_DB:
      regexStr = r'<font color="magenta">image does not exist m-[0-9a-f]+ image: comment:<\/font><br>'
      myCount = len(re.findall(regexStr, reportText))
      if myCount>5:
        reportText = re.sub(regexStr, '', reportText, count=myCount-5)
        reportText += r'<font color="magenta">image does not exist ...:<\/font><br>'
    elif task not in (Task.EXTRACTOR_TEST, Task.EXTRACTOR_RERUN, Task.DELETE_DOC, Task.EXPORT_ELN, Task.IMPORT_ELN, Task.SEND_ELAB,
                      Task.GET_ELAB, Task.SMART_ELAB):               #e.g. extractor tests work out of the box
      logging.error('Unknown task in showReport: %s', task, exc_info=True)
    showMessage(self, 'Report', reportText, image=image)


class Command(Enum):
  """ Commands used in this file """
  EXPORT     = 1
  IMPORT     = 2
  EXIT       = 3
  VIEW       = 4
  CHANGE_PG  = 6
  SYNC_SEND  = 7
  SYNC_GET   = 8
  SYNC_SMART = 9
  SCHEMA     = 10
  TEST1      = 11
  TEST2      = 12
  UPDATE     = 13
  CONFIG     = 14
  WEBSITE    = 15
  CHECK_DB   = 16
  SHORTCUTS  = 17
  RESTART    = 18
  ABOUT      = 19
  DEFINITIONS= 20
  REPOSITORY = 21



import logging, tempfile
from pasta_eln.UI.guiCommunicate import Communicate
from pasta_eln.AddOns.project_html_report import main
from .test_34_GUI_Form import getTable

def test_simple(qtbot, caplog):
  # start and wait to be up
  comm = Communicate('research')
  while comm.backendThread.worker.backend is None:
    qtbot.wait(100)

  # main part
  df = getTable(qtbot, comm, 'x0')
  projID = df[df['name']=='PASTAs Example Project']['id'].values[0]
  tempDir = tempfile.gettempdir()
  fileName = f'{tempDir}/temp.html'
  print(f'Filename {fileName}')
  main(comm, projID, None, {'fileNames':[fileName]})

  #clean shutdown
  comm.shutdownBackendThread()
  errors = [record for record in caplog.records if record.levelno >= logging.ERROR]
  assert not errors, f"Logging errors found: {[record.getMessage() for record in errors]}"



"""extract data from vendor
- tabular separated table that is used in the tensile example

THIS IS THE BASIC EXTRACTOR TUTORIAL, WHICH TEACHES
- how to create metadata conveniently using "Long name [Unit]" nomenclature
- how to create images
- code is recreated in this example, which is not ideal. But this is a tutorial and ok
"""
from io import StringIO
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def use(filePath, style={'main':''}, saveFileName=None):
  """
  Args:
    filePath (Path): full path file name
    style    (dict): supplied to guide the display / extraction style = recipe
                     main is / separated hierarchical elements parent->child
                     can contain more elements
    saveFileName (string): if given, save the image to this file-name
  Returns:
    dict: containing image, metaVendor, metaUser, style
  """
  # general information
  with open(filePath, encoding='utf-8') as fIn:
    firstLine = fIn.readline()
  data = np.loadtxt(filePath, delimiter='\t')     # use numpy to get data
  style['main'] = 'measurement/tsv/lines'

  # metadata
  if firstLine.strip() == '# strain [%], stress [MPa]':  #easy way to include extractors for multiple files into one
    maxStrain = data[:,0].max()
    strength  = data[:,1].max()
    fit = np.polyfit(data[:6,0], data[:6,1], 1)
    youngsModulus = fit[0]
    yieldIdx = np.where( np.polyval([youngsModulus, -0.2*youngsModulus], data[:,0]) >
                         data[:,1])[0][0]
    yieldStress = data[yieldIdx,1]
    direction   = filePath.stem.split('_')[1]
    # print for visual inspection
    plt.axhline(strength, color='k', linestyle=':')
    plt.axvline(maxStrain, color='k', linestyle=':')
    plt.plot(data[:8,0], np.polyval(fit, data[:8,0]), 'k-', linewidth=2)  #original fit
    plt.plot(data[:8,0], np.polyval([youngsModulus, -0.2*youngsModulus], data[:8,0]), 'r-')  #0.2% offset
    plt.plot(data[yieldIdx,0], yieldStress, 'ro')
    # save metadata
    metaUser = {'maxStrain [%]':maxStrain, 'strength [MPa]':strength, 'youngsModulus [MPa]':round(youngsModulus,4),
                'yieldStress [MPa]':yieldStress, 'direction':direction}  #DO NOT USE SPACES OR .; do use camel-case
    metaVendor = {}

  #plot
  plt.plot(data[:,0], data[:,1],'-')     # plot data using matplotlib: check the differences in plotting
  plt.xlabel(firstLine[1:].split(',')[0])
  plt.ylabel(firstLine[1:].split(',')[1])
  if firstLine.strip() == '# strain [%], stress [MPa]':
    plt.xlim(left=0)
    plt.ylim(bottom=0)
  ## plt.show()  #for fast testing

  #Save to file: each extractor should have this as it allows the user to create a high quality image
  if saveFileName is not None:
    plt.savefig(saveFileName, dpi=150, bbox_inches='tight')

  #convert image to svg image: 2d graphs are good examples of svg files
  figfile = StringIO()
  plt.savefig(figfile, format='svg')
  image = figfile.getvalue()

  # return everything
  return {'image':image, 'style':style, 'metaVendor':metaVendor, 'metaUser':metaUser}


def data(filePath, style):
  """
  Args:
    filePath (Path): full path file name
    style    (dict): supplied to guide the display / extraction style

  Returns:
    df: pandas dataframe or any other data structure
  """
  with open(filePath, encoding='utf-8') as fIn:
    firstLine = fIn.readline()
  data = np.loadtxt(filePath, delimiter='\t')     # use numpy to get data
  return pd.DataFrame(data, columns=firstLine[1:].split(','))



import logging, tempfile
from pasta_eln.UI.guiCommunicate import Communicate
from pasta_eln.UI.project import Project
from pasta_eln.AddOns.project_importCSV import main
from .test_34_GUI_Form import getTable

def verify(self):
  #Verify DB
  output = self.be.checkDB(outputStyle='text')
  output = '\n'.join(output.split('\n')[8:])
  print(output)
  self.assertNotIn('**ERROR', output, 'Error in checkDB')
  self.assertLessEqual(len(output.split('\n')), 8, 'Check db should have less than 8 almost empty lines')
  return

def test_simple(qtbot, caplog):
  # start and wait to be up
  comm = Communicate('research')
  while comm.backendThread.worker.backend is None:
    qtbot.wait(100)
  window = Project(comm)
  window.show()

  # main part
  df = getTable(qtbot, comm, 'x0')
  projID = df[df['name']=='PASTAs Example Project']['id'].values[0]
  main(comm, projID, window, {'fileNames':['tests/inputSamples.csv']})

  df2 = getTable(qtbot, comm, 'sample')
  assert 'sample A' in df2['name'].tolist(), 'Sample A incorrect'
  assert 'sample B' in df2['name'].tolist(), 'Sample B incorrect'
  assert 'sample C' in df2['name'].tolist(), 'Sample C incorrect'

  #clean shutdown
  comm.shutdownBackendThread()
  errors = [record for record in caplog.records if record.levelno >= logging.ERROR]
  assert not errors, f"Logging errors found: {[record.getMessage() for record in errors]}"



""" renders each leaf of project tree using QPaint """
import base64
import logging
from typing import Any
from PySide6.QtCore import QMargins, QModelIndex, QPoint, QRectF, QSize, Qt, Slot
from PySide6.QtGui import QColor, QPainter, QPen, QPixmap, QStaticText, QTextDocument
from PySide6.QtSvg import QSvgRenderer
from PySide6.QtWidgets import QStyledItemDelegate, QStyleOptionViewItem
from ..fixedStringsJson import DO_NOT_RENDER, defaultDataHierarchyNode
from ..textTools.handleDictionaries import doc2markdown
from ..textTools.stringChanges import markdownEqualizer
from .guiCommunicate import Communicate


class ProjectLeafRenderer(QStyledItemDelegate):
  """ ONE Renderer for all leafs of project tree using QPaint """
  def __init__(self, comm:Communicate) -> None:
    super().__init__()
    self.comm               = comm
    self.comm.backendThread.worker.beSendDoc.connect(self.onGetDoc)
    self.debugMode          = logging.root.level<logging.INFO
    self.widthImage         = self.comm.configuration['GUI']['imageWidthProject']
    self.widthContent       = self.comm.configuration['GUI']['widthContent']
    self.docTypeOffset      = self.comm.configuration['GUI']['docTypeOffset']
    self.frameSize          = self.comm.configuration['GUI']['frameSize']
    self.maxHeight          = self.comm.configuration['GUI']['maxProjectLeafHeight']
    self.lineSep            = 20
    self.penDefault         = QPen(QColor(self.comm.palette.text))
    self.penHighlight       = QPen(QColor(self.comm.palette.primary))
    self.penHighlight.setWidth(2)
    self.leafWidth          = -1
    self.docs:dict[str,Any] = {}   # docID: {'size':QSize, 'markdown':str, 'hidden':bool, 'index':QModelIndex}


  def paint(self, painter:QPainter, option:QStyleOptionViewItem, index:QModelIndex) -> None:    # type: ignore
    """
    Paint this item
    - coordinates: left, top
    - COS top left

    Args:
      painter (QPainter): painter
      option (QStyleOptionViewItem): option incl. current coordinates
      index (QModelIndex): index
    """
    data = index.data(Qt.ItemDataRole.UserRole+1)
    if not data or data['hierStack'] is None or self.comm is None:
      return
    docID   = data['hierStack'].split('/')[-1]
    name = self.docs.get(docID, {}).get('name','') or index.data(Qt.ItemDataRole.DisplayRole)
    docType = self.docs.get(docID, {}).get('type',[]) or data['docType']
    painter.setPen(self.penDefault)
    x0, y0 = option.rect.topLeft().toTuple()                                      # type: ignore[attr-defined]
    widthContent = min(self.widthContent,  \
                       int((option.rect.bottomRight()-option.rect.topLeft()).toTuple()[0]/2) )# type: ignore[attr-defined]
    docTypeOffset = min(self.docTypeOffset, \
                        int((option.rect.bottomRight()-option.rect.topLeft()).toTuple()[0]/3.5) )# type: ignore[attr-defined]
    bottomRight2nd = option.rect.bottomRight()- QPoint(self.frameSize+1,self.frameSize)# type: ignore[attr-defined]
    painter.fillRect(option.rect.marginsRemoved(QMargins(2,6,4,0)),  self.comm.palette.leafShadow)# type: ignore[attr-defined]
    if docType=='x':
      painter.fillRect(option.rect.marginsRemoved(QMargins(-2,3,8,5)), self.comm.palette.leafX)# type: ignore[attr-defined]
    else:
      painter.fillRect(option.rect.marginsRemoved(QMargins(-2,3,8,5)), self.comm.palette.leafO)# type: ignore[attr-defined]
    # header
    y = self.lineSep/2
    docTypeText= '/'.join(docType)
    if docType[0][0]=='x':
      docTypeText = self.comm.docTypesTitles['x1']['title'].lower()[:-1]
    maxCharacter = int(docTypeOffset/7.5)
    nameText = name if len(name)<maxCharacter else f'...{name[-maxCharacter+3:]}'
    if not data['gui'][0]:                                                               #Only draw first line
      staticText = QStaticText(f'<strong>{nameText} (...)</strong>')
      staticText.setTextWidth(docTypeOffset)
      painter.drawStaticText(x0, y0+y, staticText)
      painter.drawStaticText(x0+docTypeOffset, y0+y, QStaticText(docTypeText))
      return
    hiddenText = '     \U0001F441' if self.docs.get(docID, {}).get('hidden', False) else ''
    staticText = QStaticText(f'<strong>{nameText} {hiddenText}</strong>')
    staticText.setTextWidth(docTypeOffset)
    painter.drawStaticText(x0, y0+y, staticText)
    painter.drawStaticText(x0+docTypeOffset, y0+y, QStaticText(docTypeText))
    if self.debugMode:
      painter.drawStaticText(x0+700, y0+y, QStaticText(data['hierStack']))
    textDoc = QTextDocument()
    textDoc.setMarkdown(self.docs.get(docID, {}).get('markdown',''))
    painter.translate(QPoint(x0-3, y0+y+15))
    self.drawTextDocument(painter, textDoc, int(self.maxHeight-6*self.frameSize))
    painter.translate(-QPoint(x0-3, y0+y+15))
    # right side
    if self.docs.get(docID, {}).get('content','') and not self.docs.get(docID, {}).get('image',''):
      textDoc = QTextDocument()
      textDoc.setMarkdown(self.docs.get(docID, {}).get('content',''))
      textDoc.setTextWidth(widthContent)
      width:int = textDoc.size().toTuple()[0]                                                   # type: ignore
      topLeftContent = option.rect.topRight() - QPoint(width+self.frameSize-2,-self.frameSize)# type: ignore[attr-defined]
      painter.translate(topLeftContent)
      self.drawTextDocument(painter, textDoc, int(self.maxHeight-3*self.frameSize))
      topLeftContent = option.rect.topRight() - QPoint(width+self.frameSize-2,-self.frameSize)# type: ignore[attr-defined]
      painter.translate(-topLeftContent)
    if self.docs.get(docID, {}).get('image',''):
      if self.docs.get(docID, {}).get('image','').startswith('data:image/'):
        pixmap = self.imageFromDoc({'image':self.docs.get(docID, {}).get('image','')})
        width2nd = min(self.widthImage, pixmap.width()+self.frameSize)
        topLeft2nd     = option.rect.topRight()   - QPoint(width2nd+self.frameSize+1,-self.frameSize)# type: ignore[attr-defined]
        painter.drawPixmap(topLeft2nd, pixmap)
      elif self.docs.get(docID, {}).get('image','').startswith('<?xml'):
        topLeft2nd     = option.rect.topRight()   - QPoint(self.widthImage+self.frameSize+1,-self.frameSize)# type: ignore[attr-defined]
        image = QSvgRenderer(bytearray(self.docs.get(docID, {}).get('image',''), encoding='utf-8'))
        image.render(painter,    QRectF(topLeft2nd, bottomRight2nd))
    return


  def sizeHint(self, option:QStyleOptionViewItem, index:QModelIndex) -> QSize:                  # type: ignore
    """
    determine size of this leaf
    """
    if not index or not index.data(Qt.ItemDataRole.UserRole+1):
      return QSize()
    hierStack = index.data(Qt.ItemDataRole.UserRole+1)['hierStack']
    if hierStack is None or self.comm is None:
      return QSize()
    if not index.data(Qt.ItemDataRole.UserRole+1)['gui'][0]:              # only show the headline, no details
      return QSize(400, self.lineSep*2)
    docID   = hierStack.split('/')[-1]
    if docID not in self.docs:
      self.leafWidth = min(self.widthContent,
                           int((option.rect.bottomRight()-option.rect.topLeft()).toTuple()[0]/2) )# type: ignore[attr-defined]
      self.docs[docID] = {'size':QSize(400, 30), 'markdown':'', 'hidden':False, 'index':index}
      self.comm.uiRequestDoc.emit(docID)
    return self.docs[docID].get('size', QSize(400,self.maxHeight))


  @Slot(str)
  def onGetDoc(self, doc:dict[str,Any]) -> None:
    """ Slot to handle the document received from backend
    Args:
      doc (dict): document
    """
    guiStyle = self.comm.configuration['GUI']
    if not doc or doc['id'] not in self.docs:
      return
    logging.debug('Renderer: onGetDoc %s %s %s', doc['id'], doc.get('type',[]), doc.get('hierStack',''))
    # ... after deleting project, its items cannot be found and it would give many false negatives
    if doc['type'][0] not in self.comm.docTypesTitles:
      dataHierarchyNode = defaultDataHierarchyNode
    else:
      dataHierarchyNode = self.comm.dataHierarchyNodes[doc['type'][0]]
    textDoc = QTextDocument()
    markdownStr = doc2markdown(doc, DO_NOT_RENDER, dataHierarchyNode, self)
    textDoc.setMarkdown(markdownStr)
    textDoc.setTextWidth(self.leafWidth)
    heightDetails = int(textDoc.size().toTuple()[1])+guiStyle['frameSize']+20# type: ignore
    heightRightSide = -1
    if 'content' in doc:
      textDoc.setMarkdown(doc['content'])
      heightRightSide = int(textDoc.size().toTuple()[1])                                        # type: ignore
    elif 'image' in doc and doc['image']:
      if doc['image'].startswith('data:image/'):
        pixmap = self.imageFromDoc(doc)
        heightRightSide = pixmap.height()+2*guiStyle['frameSize']
      else:
        heightRightSide = int(guiStyle['imageWidthProject']*3/4+2*guiStyle['frameSize'])
    self.docs[doc['id']]['size']    = QSize(400, min(max(heightDetails,heightRightSide), guiStyle['maxProjectLeafHeight']))
    self.docs[doc['id']]['hidden']  = any(b for b in doc['branch'] if False in b['show'])
    self.docs[doc['id']]['markdown']= markdownStr
    self.docs[doc['id']]['name']    = doc['name']
    self.docs[doc['id']]['type'] = doc['type']
    self.docs[doc['id']]['content'] = markdownEqualizer(doc['content']) if 'content' in doc else ''
    self.docs[doc['id']]['image']   = doc.get('image','')
    self.sizeHintChanged.emit(self.docs[doc['id']]['index'])


  def drawTextDocument(self, painter:QPainter, textDoc:QTextDocument, yMax:int) -> None:
    """ Draw text document

    Args:
      painter (QPainter): painter
      textDoc (QTextDocument): text document
      yMax (int): maximum height of document in surrounding frame
    """
    width:int  = textDoc.size().toTuple()[0]                                                    # type: ignore
    height:int = textDoc.size().toTuple()[1]                                                    # type: ignore
    textDoc.drawContents(painter, QRectF(0, 0, width, yMax))
    if height > yMax+self.frameSize:
      painter.setPen(self.penHighlight)
      painter.drawLine(self.frameSize, yMax+self.frameSize, width-self.frameSize, yMax+self.frameSize)
      painter.setPen(self.penDefault)
    return


  def imageFromDoc(self, doc:dict[str,Any]) -> QPixmap:
    """ Create image from image in doc

    Args:
      doc (dict): document

    Returns:
      QPixmap: image
    """
    result = QPixmap()
    result.loadFromData(base64.b64decode(doc['image'][22:]))
    result = result.scaledToWidth(self.widthImage)
    if result.height()>self.maxHeight:
      result = result.scaledToHeight(self.maxHeight-self.frameSize*2)
    return result



""" Displays a scrollable grid of images (PNG, JPG, SVG) """
import logging
from typing import Any
from PySide6.QtCore import QByteArray, Qt, Signal, Slot
from PySide6.QtGui import QImage, QMouseEvent, QPixmap, QStandardItemModel
from PySide6.QtSvgWidgets import QSvgWidget
from PySide6.QtWidgets import QGridLayout, QPushButton, QScrollArea, QVBoxLayout, QWidget
from .guiCommunicate import Communicate

IMG_SIZE = 300

class ClickableFrame(QPushButton):
  """
  A QPushbutton subclass that emits signals when clicked or double-clicked
  """
  clicked = Signal(str)
  doubleClicked = Signal(str)

  def __init__(self, docID: str):
    """
    Initialize

    Args:
      docID: The unique identifier for the image
    """
    super().__init__()
    self.docID = docID
    self.setFixedSize(IMG_SIZE, IMG_SIZE)
    self.setStyleSheet('border: none;')


  def mousePressEvent(self, event:QMouseEvent) -> None:
    """
    Handles mouse press events. Emits 'clicked' signal on left-button press

    Args:
      event: The QMouseEvent
    """
    if event.button() == Qt.LeftButton:                                           # type: ignore[attr-defined]
      self.clicked.emit(self.docID)
    return


  def mouseDoubleClickEvent(self, event:QMouseEvent) -> None:
    """
    Handles mouse double-click events. Emits 'doubleClicked' signal on left-button double-click

    Args:
      event: The QMouseEvent
    """
    if event.button() == Qt.LeftButton:                                           # type: ignore[attr-defined]
      self.doubleClicked.emit(self.docID)
    return


class ImageGallery(QWidget):
  """
  A widget that displays a scrollable grid of images (PNG, JPG, SVG)
  Images can be clicked or double-clicked to trigger actions
  """
  def __init__(self, comm:Communicate):
    """
    Initializes the ImageGallery

    Args:
      comm: The communication object used to interact with the backend and other UI components
    """
    super().__init__()
    self.comm = comm
    self.comm.backendThread.worker.beSendDoc.connect(self.onGetData)
    self.data:dict[str,tuple[int,int]] = {}
    self.model:QStandardItemModel | None = None
    layout = QVBoxLayout(self)
    scrollArea = QScrollArea(self)
    scrollArea.setWidgetResizable(True)
    scrollContent = QWidget()
    self.gridL = QGridLayout(scrollContent)
    scrollArea.setWidget(scrollContent)
    layout.addWidget(scrollArea)

  def updateGrid(self, model:QStandardItemModel) -> None:
    """
    Args:
      model: The data model containing information about the images to display
             The first column of each row is expected to have an 'accessibleText'
    """
    self.data = {}
    self.model = model
    row, col = 0, 0
    for idx in range(self.model.rowCount()):
      docID = model.itemFromIndex(model.index(idx,0)).accessibleText()
      self.data[docID] = (row, col)
      col += 1
      if col >= 4:                                                                        # Assuming 4 columns
        col = 0
        row += 1
    # Clear existing widgets from the grid
    while self.gridL.count():
      child = self.gridL.takeAt(0)
      if child.widget():
        child.widget().deleteLater()
    return


  @Slot(dict)
  def onGetData(self, doc:dict[str,Any]) -> None:
    """
    Populates or updates the image grid based on the provided model

    Clears the existing grid and then iterates through the model to fetch
    and display images. It handles both SVG and raster image formats (like PNG, JPG)

    Args:
      doc (dict[str, Any]): A dictionary containing the document data.
    """
    if 'id' in doc and doc['id'] in self.data:
      if not doc or 'image' not in doc:
        # Handle cases where image data might be missing or malformed
        logging.warning('Image data not found or malformed for docID: %s', doc['id'])
        return
      image = doc['image']
      row, col = self.data[doc['id']]

      if image.startswith('<?xml version="1.0"'):
        # Wrap QSvgWidget in a Clickable subclass of QPushButton
        button = ClickableFrame(doc['id'])
        button.setFixedSize(IMG_SIZE, IMG_SIZE)
        svgWidget = QSvgWidget()
        svgWidget.renderer().load(bytearray(image, encoding='utf-8'))
        # Calculate aspect ratio for SVG
        svg_width = svgWidget.renderer().defaultSize().width()
        svg_height = svgWidget.renderer().defaultSize().height()

        if svg_width == 0 or svg_height == 0:                                         # Avoid division by zero
          new_width, new_height = IMG_SIZE -4, IMG_SIZE -4
        elif svg_height > svg_width:
          new_width = int(float(svg_width) / float(svg_height) * (IMG_SIZE - 4))
          new_height = IMG_SIZE - 4
        else:
          new_width = IMG_SIZE - 4
          new_height = int(float(svg_height) / float(svg_width) * (IMG_SIZE - 4))
        svgWidget.setMaximumSize(new_width, new_height)

        layout = QVBoxLayout(button)
        layout.addWidget(svgWidget)
        layout.setAlignment(Qt.AlignCenter)                                       # type: ignore[attr-defined]
        button.clicked.connect(lambda checked=False, docID=doc['id']: self.imageClicked(docID))
        button.doubleClicked.connect(self.image2Clicked)
        self.gridL.addWidget(button, row, col)
      elif 'base64,' in image:                                               # Basic check for base64 data URI
        try:
          header, base64_data = image.split(',', 1)
          # Extract image type from header (e.g., "data:image/png;base64")
          img_type_part = header.split(';')[0].split('/')[-1]
          imageType = img_type_part.upper()

          byteArr = QByteArray.fromBase64(bytearray(base64_data, encoding='utf-8'))
          imageW = QImage()
          # Ensure format string is clean (e.g. PNG, JPEG)
          fmt = imageType.replace(';', '')
          if imageW.loadFromData(byteArr, format=fmt):
            pixmap = QPixmap.fromImage(imageW).scaled(IMG_SIZE,IMG_SIZE,Qt.KeepAspectRatio,Qt.SmoothTransformation)# type: ignore[attr-defined]
            button = ClickableFrame(doc['id'])
            button.setIcon(pixmap)
            button.setAlignment(Qt.AlignCenter)                                   # type: ignore[attr-defined]
            button.clicked.connect(self.imageClicked)
            button.doubleClicked.connect(self.image2Clicked)
            self.gridL.addWidget(button, row, col)
          else:
            logging.warning('Could not load image data for docID: %s with format %s', doc['id'], fmt)
        except Exception as e:
          logging.warning('Error processing image for docID %s: %s', doc['id'], e)
      else:
        logging.warning('Image for docID is not in expected base64 format. %s', doc['id'])
    return


  def imageClicked(self, docID:str) -> None:
    """
    Handles single-click events on an image

    Args:
      docID: The unique identifier of the clicked image/document
    """
    self.comm.changeDetails.emit(docID)
    return


  def image2Clicked(self, docID:str) -> None:
    """
    Handles double-click events on an image

    Args:
      docID: The unique identifier of the double-clicked image/document
    """
    self.comm.formDoc.emit({'id':docID})
    return



#!/usr/bin/python3
"""TEST the project view: drag drop randomly items around """
import logging, warnings, random
from pathlib import Path
from anytree import PreOrderIter
from PySide6.QtCore import QModelIndex, QEventLoop                         # pylint: disable=no-name-in-module
from pasta_eln.UI.project import Project
from pasta_eln.UI.guiCommunicate import Communicate
from pasta_eln.backendWorker.worker import Task


def test_simple(qtbot, caplog):
  """
  main function
  """
  # initialization: create database, destroy on filesystem and database and then create new one
  warnings.filterwarnings('ignore', message='numpy.ufunc size changed')
  warnings.filterwarnings('ignore', message='invalid escape sequence')
  warnings.filterwarnings('ignore', category=ResourceWarning, module='PIL')
  warnings.filterwarnings('ignore', category=ImportWarning)
  logPath = Path.home()/'pastaELN.log'
  logging.basicConfig(filename=logPath, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                      datefmt='%m-%d %H:%M:%S')   #This logging is always info, since for installation only
  for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib.font_manager']:
    logging.getLogger(package).setLevel(logging.WARNING)
  logging.info('Start 03 test')

  # start app and load project
  comm = Communicate('research')
  window = Project(comm)
  qtbot.addWidget(window)
  while comm.backendThread.worker.backend is None:
    qtbot.wait(100)
  projID = comm.backendThread.worker.backend.output('x0').split('|')[-2].strip()  #for testing purposes
  window.change(projID,'')

  choices = random.choices(range(100), k=16)
  # choices =
  print(f'Current choice: [{",".join([str(i) for i in choices])}]')
  # start iteration
  for epoch in range(4):
    print(f'{"*"*40}\nStart drag-drop {epoch}\n{"*"*40}')
    def recursiveRowIteration(index:QModelIndex) -> list[QModelIndex]:
      item  = window.tree.model().itemFromIndex(index)
      allParentIdx = [index] if item is not None and item.data() is not None and item.data()['docType']==['x1'] else []
      for subRow in range(window.tree.model().rowCount(index)):
        subIndex = window.tree.model().index(subRow,0, index)
        allParentIdx += recursiveRowIteration(subIndex)
      return allParentIdx
    qtbot.wait(1000)
    allParentIdx = recursiveRowIteration( window.tree.model().index(-1,0))
    validChoices     = [i for i in allParentIdx if window.tree.model().rowCount(i)>0 ]
    sourceParentIdx  = validChoices[choices.pop(0)%len(validChoices)]
    sourceParentItem = window.tree.model().itemFromIndex(sourceParentIdx)
    validChoices     = range(window.tree.model().rowCount(sourceParentIdx))
    sourceChildRow   = validChoices[choices.pop(0)%len(validChoices)]
    sourceItem       = sourceParentItem.takeChild(sourceChildRow)
    if sourceItem.data() is None:
      continue
    validChoices     = [window.tree.model().itemFromIndex(i) for i in allParentIdx
                        if window.tree.model().itemFromIndex(i).data() is not None and
                          window.tree.model().itemFromIndex(i).data()['hierStack'].split('/')[-1] != sourceItem.data()['hierStack'].split('/')[-1] ]
    targetParent     = validChoices[choices.pop(0)%len(validChoices)]
    validChoices     = range(targetParent.rowCount()+1)
    targetChildRow   = validChoices[choices.pop(0)%len(validChoices)]
    print('  ',sourceItem.data(),'->\n  ', targetParent.data(),'   child', targetChildRow)
    targetParent.setChild(targetChildRow, sourceItem)
    verify(comm, projID, epoch)

  # close everything
  print(f'{"*"*40}\nEND TEST 03\n{"*"*40}')
  comm.shutdownBackendThread()

  errors = [record for record in caplog.records if record.levelno >= logging.ERROR]
  assert not errors, f"Logging errors found: {[record.getMessage() for record in errors]}"
  return


def verify(comm, projID, epoch): # Output hierarchy and verify DB
  # to communicate with backend
  loop = QEventLoop()
  def hierarchyCallback(hierarchy):
    print(f'{"*"*40}\nHierarchy after drag-drop {epoch}\n{"*"*40}')
    print(''.join('  '*node.depth + node.name + ' | ' + '/'.join(node.docType) + (f' | {node.id}') +'\n'
                   for node in PreOrderIter(hierarchy)))
    loop.quit()
  def checkDBCallback(_, output):
    print(f'{"*"*40}\nCheckDB after drag-drop {epoch}\n{"*"*40}')
    print(output)
    output = '\n'.join(output.split('\n')[8:])
    assert '**ERROR' not in output, 'Error in checkDB'
    loop.quit()
  comm.backendThread.worker.beSendHierarchy.connect(hierarchyCallback)
  comm.backendThread.worker.beSendTaskReport.connect(checkDBCallback)
  comm.uiRequestHierarchy.emit(projID, True)
  loop.exec()
  comm.uiRequestTask.emit(Task.CHECK_DB, {'style':'text'})
  loop.exec()
  comm.backendThread.worker.beSendHierarchy.disconnect(hierarchyCallback)
  comm.backendThread.worker.beSendTaskReport.disconnect(checkDBCallback)
  return



""" Main methods that start the gui """
import logging
import sys
import traceback
from pathlib import Path
from PySide6.QtCore import QCoreApplication
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QApplication, QMessageBox
from pasta_eln import __version__
from .miscTools import testNewPastaVersion
from .UI.guiCommunicate import Communicate
from .UI.mainWindow import MainWindow


def mainGUI(projectGroup:str='') -> tuple[QCoreApplication | None, MainWindow]:
  """  Main method and entry point for commands

  Args:
      projectGroup (str): project group to load

  Returns:
    MainWindow: main window
  """
  # logging has to be started first
  log_path = Path.home() / 'pastaELN.log'
  #  old versions of basicConfig do not know "encoding='utf-8'"
  logging.basicConfig(filename=log_path, level=logging.INFO, format='%(asctime)s|%(levelname)s|%(filename)s:%(lineno)d:%(message)s',
                      datefmt='%m-%d %H:%M:%S')
  for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib','pudb']:
    logging.getLogger(package).setLevel(logging.WARNING)
  logging.info('Start PASTA GUI')
  # remainder
  if not QApplication.instance():
    application = QApplication().instance()
  else:
    application = QApplication.instance()
  comm = Communicate(projectGroup=projectGroup)
  mainWindow = MainWindow(comm)
  logging.getLogger().setLevel(getattr(logging, comm.configuration.get('GUI',{'loggingLevel':'DEBUG'})['loggingLevel']))
  if mainWindow.comm.palette is not None and application is not None:
    mainWindow.comm.palette.setTheme(application)                                    # type: ignore [arg-type]
  import qtawesome as qta  # qtawesome and matplot cannot coexist
  if not isinstance(qta.icon('fa5s.times'), QIcon):
    logging.error('qtawesome: could not load. Likely matplotlib is included and can not coexist.')
    print('qtawesome: could not load. Likely matplotlib is included and can not coexist.')
  # end test coexistence
  # Test for new version
  if comm.configuration.get('GUI', {'checkForUpdates':'No'})['checkForUpdates']=='Yes' and not testNewPastaVersion(False):
    button = QMessageBox.question(mainWindow, 'Update?', 'There is a new PASTA-ELN version available. Do you want to update?',
                                  QMessageBox.StandardButton.No, QMessageBox.StandardButton.Yes)
    if button == QMessageBox.StandardButton.Yes:
      testNewPastaVersion(update=True)
  return application, mainWindow


def startMain(projectGroup:str='') -> None:
  """
  Main function to start GUI. Extra function is required to allow starting in module fashion

  Args:
    projectGroup (str): project group to load
  """
  try:
    app, window = mainGUI(projectGroup=projectGroup)
    window.show()
    if app:
      app.exec()
  except Exception as e:
    traceback.print_exc()
    logging.error('Error in mainGUI: %s', e, exc_info=True)


# called by python3 -m pasta_eln.gui
if __name__ == '__main__':
  startMain(sys.argv[1] if len(sys.argv)>1 else '')



""" Communication class that sends signals between widgets and the backend worker"""
import logging
import os
import sys
from pathlib import Path
from typing import Any
from PySide6.QtCore import QObject, Signal, Slot
from ..backendWorker.worker import BackendThread, Task
from ..miscTools import getConfiguration
from .palette import Palette
from .waitDialog import WaitDialog, Worker


class Communicate(QObject):
  """ Communication class that sends signals between widgets and the backend worker"""
  # General concept of the signals
  # - signals are defined either
  #   - here (signals / data send within the GUI or send data from GUI to backend worker)
  #   - in backendWorker/worker.py (signals = data send from backend worker to GUI)
  # - try to connect straight to those signals
  #   - only data that is common to two or more widgets is saved here here

  # BE SPECIFIC ABOUT WHAT THIS ACTION DOES
  # signals request a change within the UI elements:
  changeSidebar      = Signal(str)       # redraw sidebar after hide/show of project in table, focus on this projectID
  changeTable        = Signal(str, str)    # send doctype,projectID from sidebar to main-table
                                           #      can also be used for hiding the details on right side if nothing to show
  changeDetails      = Signal(str)       # send docID from main-table to details
                                         #      docID (str): document-id; ''=draw nothing; 'redraw' implies redraw
  changeProject      = Signal(str, str)  # send docID,projectID from sidebar or main-table to projects
  stopSequentialEdit = Signal()          # in sequential edit, stop if there is a cancel
  # send data or data-request to backend
  commSendConfiguration = Signal(dict, str)     # send configuration and project-group-name to backend
  uiRequestDataHierarchy= Signal(str)           # get all entries in the data hierarchy for this docType
  uiRequestTable        = Signal(str, str, bool)# table: send docType, projectID, showAll to backend to get table
  uiRequestHierarchy    = Signal(str, bool)     # send project ID to backend
  uiRequestDoc          = Signal(str)           # request doc
  uiRequestTask         = Signal(Task, dict)    # request to execute a task
  uiSendSQL             = Signal(list)          # request to execute SQL commands directly
  # signals that are emitted from this comm that data changed
  docTypesChanged    = Signal()          # redraw main window, e.g. after change of docType titles

  # unclear
  formDoc            = Signal(dict)      # send doc from details to new/edit dialog: dialogForm
  testExtractor      = Signal()          # execute extractorTest in widgetDetails

  def __init__(self, projectGroup:str=''):
    super().__init__()
    self.waitDialog            = WaitDialog()
    self.worker:Worker|None    = None
    self.palette                                = Palette(None, 'none')               #reset to real one later
    self.configuration, self.projectGroup = getConfiguration(projectGroup)
    if not self.configuration:
      return
    self.basePath = Path(self.configuration['projectGroups'][self.projectGroup]['local']['path'])
    self.addOnPath = self.configuration['projectGroups'][self.projectGroup]['addOnDir']
    logging.info('Initial python path: %s', sys.path)
    logging.info('Backend initialized with basePath %s and prepend addOnPath %s', self.basePath, self.addOnPath)

    # Data storage for all widgets
    self.docTypesTitles:dict[str,dict[str,str]] = {}# docType: {'title':title,'icon':icon,'shortcut':shortcut}
    self.dataHierarchyNodes:dict[str,list[Any]] = {}
    self.projectID                              = ''

    if self.configuration:
      # Backend worker thread
      self.backendThread = BackendThread(self)
      # connect backend worker to configuration signals: send GUI->backend
      #   has to be here, else worker needs comm which has to be passed through thread, is uninitialized, ...)
      # connect backend worker SLOTS to GUI signals: group B
      self.commSendConfiguration.connect(self.backendThread.worker.initialize)
      self.uiRequestTable.connect(self.backendThread.worker.returnTable)
      self.uiRequestHierarchy.connect(self.backendThread.worker.returnHierarchy)
      self.uiRequestDoc.connect(self.backendThread.worker.returnDoc)
      self.uiRequestTask.connect(self.backendThread.worker.returnTaskReport)
      self.uiSendSQL.connect(self.backendThread.worker.executeSQL)

      # connect GUI SLOTS to backend worker signals: group C
      self.backendThread.worker.beSendDataHierarchyNode.connect(self.onGetDataHierarchyNode)
      self.backendThread.worker.beSendDocTypes.connect(self.onGetDocTypes)

      # connect waiting dialog
      self.uiRequestTask.connect(self.progressWindow)
      self.backendThread.worker.beSendTaskReport.connect(self.waitDialog.hide)

      # start thread now that everything is linked up
      self.backendThread.start()
      self.start()


  @Slot(dict)
  def onGetDocTypes(self, data: dict[str, dict[str, str]]) -> None:
    """ Handle data received from backend worker
    Args:
      data (dict): dictionary with docType: {'title':title,'icon':icon,'shortcut':shortcut}
    """
    self.docTypesTitles = data
    self.docTypesChanged.emit()


  @Slot(str, list)
  def onGetDataHierarchyNode(self, docType:str, data:list[Any]) -> None:
    """ Handle data received from backend worker
    Args:
      docType (str): document type
      data (list): list of nodes in the hierarchy for this docType
    """
    self.dataHierarchyNodes[docType] = data


  def start(self, projectGroup:str='') -> None:
    """
    Start the backend thread

    Args:
      projectGroup (str): project group to use, if empty use the one from initialization
    """
    self.commSendConfiguration.emit(self.configuration, projectGroup or self.projectGroup)
    self.changeTable.emit('x0', '')                                      # show project table, without details


  def shutdownBackendThread(self) -> None:
    """
    Shutdown backend thread
    """
    if self.backendThread is not None:
      self.backendThread.quit()
      self.backendThread.wait()
      del self.backendThread
    self.waitDialog.close()


  def progressWindow(self, task:Task, _:dict[str,Any]) -> None:
    """ Show a progress window and execute function
    Args:
      task (Task): task to execute
    """
    if task.msgWaitDialog == '' or 'PYTEST_CURRENT_TEST' in os.environ:
      return
    self.waitDialog.text.setMarkdown(task.msgWaitDialog)
    self.waitDialog.text.setFixedHeight(30)
    self.waitDialog.setFixedHeight(100)
    self.waitDialog.progressBar.setRange(0, 0)  # Indeterminate
    self.waitDialog.show()



""" Functions that modify strings into the appropriate format """
import logging
import re
import traceback
from pathlib import Path
from typing import Any
from ..miscTools import Bcolors


def outputString(fmt:str='print', level:str='info', message:str='') -> str:
  """ Output a message into different formats:
    - print: print to stdout
    - logging; log to file
    - text: return text string (supersedes html)
    - html: return html string https://doc.qt.io/qtforpython/overviews/richtext-html-subset.html#supported-html-subset
    - else: no output
    - formats can be union ('print,text')
  """
  prefixes = {'h2':f'{Bcolors.UNDERLINE}\n*** ','bold':f'{Bcolors.BOLD}\n*** ', \
              'perfect':f'{Bcolors.OKGREEN}', 'ok':f'{Bcolors.OKBLUE}', 'unsure':f'{Bcolors.HEADER}',\
              'warning':f'{Bcolors.WARNING}**Warning','error':f'{Bcolors.FAIL}**ERROR '}
  if level=='info':
    txtOutput = message.strip()+'\n'
  elif level in prefixes:
    txtOutput = f'{prefixes[level]} {message}'
    txtOutput+= ' ***' if '***' in prefixes[level] else ''
    txtOutput+= f'{Bcolors.ENDC}\n'
  else:
    logging.error('Level not in prefixes %s',level, exc_info=True)
  # depend on format
  if 'print' in fmt:
    print(txtOutput)
  if 'logging' in fmt and level in {'info', 'warning', 'error'}:
    getattr(logging,level)(message)
  if 'text' in fmt:
    return txtOutput
  if fmt=='html':
    colors = {'info':'black','error':'red','warning':'orangered','perfect':'green','ok':'blue','unsure':'magenta'}
    if level[0]=='h':
      return f'<{level}>{message}</{level}>'
    if level not in colors:
      logging.error('Wrong level %s', level, exc_info=True)
      return ''
    return f'<font color="{colors[level]}">' + message.replace('\n', '<br>') + '</font><br>'
  return ''


def tracebackString(log:bool=False, docID:str='') -> str:
  """ Create a formatted string of traceback

  Args:
    log (bool): write to logging
    docID (str): docID used in comment

  Returns:
    str: | separated string of call functions
  """
  tracebackList = [i.split('\n')[0] for i in traceback.format_stack()[2:-2] if 'pasta_eln' in i]#skip first and last and then filter only things with pasta_eln
  reply = '|'.join([item.split('/')[-1].strip() for item in tracebackList])#| separated list of stack excluding last
  reply = reply.replace('",','')
  if log:
    logging.info(' traceback %s %s', docID, reply)
  return reply


def markdownEqualizer(text:str) -> str:
  """
  Create a markdown that well balanced with regard to font size, etc

  Args:
    text (str): input string

  Returns:
    str: output str
  """
  if isinstance(text, tuple):
    text=text[0]
  return re.sub(r'(^|\n)(#+)', r'\1##\2', text.strip())


def tuple2html(key:str, value:tuple[str,str,str,str]) -> tuple[str,str]:
  """
  Convert tuple to html string

  Args:
    key   (str): key/label to add
    value (tuple): tuple to convert

  Returns:
    str,str: key, html string
  """
  key = key if value[2] is None or value[2]=='' else value[2]
  valueString = f'{value[0]} {value[1]}'
  valueString = valueString if value[3] is None or value[3]=='' else \
                f'{valueString}&nbsp;<b><a href="{value[3]}">&uArr;</a></b>'
  return key, valueString


def camelCase(text:str) -> str:
  """
  Produce camelCase from normal string
  - file names abcdefg.hij are only replaced spaces

  Args:
     text (str): string

  Returns:
    str: camel case of that string: CamelCaseString
  """
  if re.match(r'^[\w-]+\.[\w]+$', text):
    return text.replace(' ','_')
  return re.sub(r'(_|-)+', ' ', text).title().replace(' ','').replace('*','')


def createDirName(doc:dict[str,Any], thisChildNumber:int, parentDir:Path) -> str:
  """ create directory-name by using camelCase and a prefix

  Args:
      doc (dict): document with all information
      thisChildNumber (int): number of myself
      parentDir (Path): parent directory where the new directory should be created

  Returns:
    string: directory name with leading number
  """
  name = camelCase(doc['name']) if doc['type'][0] == 'x0' else \
         f'{thisChildNumber:03d}_{camelCase(doc["name"])}'
  if 'branch' in doc and name in [i['path'].split('/')[-1] for i in doc['branch']]:#only change if not the same name as before
    logging.debug('createDirName: %s used', name)
    return name
  idx = 0
  nameTest = name
  while (parentDir/name).exists():
    idx += 1
    nameTest = f'{name}_{idx:02d}'
  logging.debug('createDirName: %s created', nameTest)
  return nameTest



""" Backend worker thread for separating GUI and backend operations
CONNECT TO ALL THESE SIGNALS IN COMMUNICATE and UI
"""
import json
import logging
import os
import shutil
import tempfile
import time
from collections import Counter
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Optional
import pandas as pd
from anytree import Node
from PySide6.QtCore import QObject, QThread, Signal, Slot
from PySide6.QtWidgets import QMessageBox
from ..miscTools import flatten
from ..textTools.stringChanges import createDirName
from .backend import Backend
from .dataverse import DataverseClient
from .elabFTWsync import MERGE_LABELS, Pasta2Elab
from .inputOutput import exportELN, importELN
from .zenodo import ZenodoClient

waitTimeBeforeSendingFirstMessage = 0.1 #ensure all UI elements are up

class Task(Enum):
  """ Tasks that can be used in BackendWorker """
  ADD_DOC        = (1 , '')                                        #keys: hierStack, docType, doc
  EDIT_DOC       = (2 , '')                                        #keys: doc, newProjID
  MOVE_LEAVES    = (3 , '')                                        #keys: docID, stackOld, stackNew, childOld, childNew
  DROP_EXTERNAL  = (4 , 'Including drag&drop files and folders:')  #keys: docID, files, folders
  HIDE_SHOW      = (5 , '')                                        #keys: docID
  SET_GUI        = (6 , '')                                        #keys: docID, gui
  DELETE_DOC     = (7 , '')                                        #keys: docID
  SCAN           = (8 , 'Scanning disk for new data:')             #keys: docID
  SEND_TBL_COLUMN= (9 , '')                                        #keys: docType, newList
  EXTRACTOR_TEST = (10, 'Testing extractor:')                      #keys: fileName, style, recipe, saveFig
  EXTRACTOR_RERUN= (11, 'Rerun extractors:')                       #keys: docIDs, recipe
  SEND_ELAB      = (12, 'Sending data to elabFTW server:')         #keys: projGroup
  GET_ELAB       = (13, 'Getting data from elabFTW server:')       #keys: projGroup
  SMART_ELAB     = (14, 'Syncing with elabFTW server:')            #keys: projGroup
  EXPORT_ELN     = (15, 'Exporting to .eln:')                      #keys: fileName, projID, docTypes
  IMPORT_ELN     = (16, 'Importing an .eln file:')                 #keys: fileName, projID
  SEND_REPOSITORY= (17, 'Sending data to repository:')             #keys: projID, docTypes,repositories,metadata,uploadZenodo
  CHECK_DB       = (18, 'Checking database integrity:')            #keys: style
  OPEN_EXTERNAL  = (19, '')                                        #keys: docID

  def __init__(self, num:int, msgWaitDialog:str='') -> None:
    """ Initialize the task with a number and an optional message for the wait dialog
    Args:
      num (int): Task number
      msgWaitDialog (str): Message to show in the wait dialog
    """
    self.msgWaitDialog = msgWaitDialog


class BackendWorker(QObject):
  """
  Backend worker that runs in a separate thread to handle all backend operations
  """
  # Signals to send data back to GUI
  beSendDocTypes          = Signal(dict)           # Send processed data back
  beSendDataHierarchyNode = Signal(str,list)       # Send data hierarchy nodes
  beSendDataHierarchyAll  = Signal(list)           # Send all entries for this docType
  beSendTable             = Signal(pd.DataFrame, str)   # all tables
  beSendHierarchy         = Signal(Node, dict)
  beSendDoc               = Signal(dict)
  beSendTaskReport        = Signal(Task, str, str, str)       # task, report, image, path
  beSendSQL               = Signal(str, pd.DataFrame)

  def __init__(self) -> None:
    """ Initialize the backend worker """
    super().__init__()
    self.backend: Optional[Backend] = None


  @Slot(dict,str)
  def initialize(self, configuration:dict[str,Any], projectGroupName:str) -> None:
    """ Initialize the backend worker with the given configuration
    Args:
      configuration (dict): Configuration dictionary with database and other settings
      projectGroupName (str): Name of the project group to initialize
    """
    self.backend = Backend(configuration, projectGroupName)
    docTypesTitlesIcons = {k:{'title':v} for k,v in self.backend.db.dataHierarchy('','title')}
    for k,v in self.backend.db.dataHierarchy('','icon'):
      docTypesTitlesIcons[k]['icon'] = v
    for k,v in self.backend.db.dataHierarchy('','shortcut'):
      docTypesTitlesIcons[k]['shortcut'] = v
    time.sleep(waitTimeBeforeSendingFirstMessage)
    self.beSendDocTypes.emit(docTypesTitlesIcons)
    for docType in docTypesTitlesIcons:
      self.beSendDataHierarchyNode.emit(docType, self.backend.db.dataHierarchy(docType, 'meta'))

  @Slot(str)
  def returnDataHierarchyRow(self, docType:str) -> None:
    """ Return a data hierarchy row for the given docType
    Args:
      docType (str): Document type to return data hierarchy for
    """
    if self.backend is not None:
      self.beSendDataHierarchyAll.emit(self.backend.db.dataHierarchy(docType, '*'))


  @Slot(str, str, bool)
  def returnTable(self, docType:str, projID:str, showAll:bool) -> None:
    """ Return a view from the database
    Args:
      docType (str): Document type to return
      projID (str): Project ID to get the view for
      showAll (bool): Whether to return all items or only the non-hidden ones"""
    if docType and self.backend is not None:
      if docType=='_tags_':
        path = 'viewIdentify/viewTags'
      else:
        path = f'viewDocType/{docType}'
      path += 'All' if showAll else ''
      logging.debug('returnTable %s %s %s %s', docType, projID, showAll, path)
      data = self.backend.db.getView(path, startKey=projID)
      self.beSendTable.emit(data, docType)


  @Slot(str, bool)
  def returnHierarchy(self, projID:str, showAll:bool) -> None:
    """ Return a hierarchy
    Args:
      projID (str): Project ID to get the hierarchy for
      showAll (bool): Whether to return all items or only the non-hidden ones
    """
    if self.backend is not None and projID:#TODO: during test_13 for some reason the projID is empty, not sure why
      hierarchy, error = self.backend.db.getHierarchy(projID, allItems=showAll)
      if error:
        hierarchy = Node('__ERROR_in_getHierarchy__')
      projDoc = self.backend.db.getDoc(projID)
      logging.debug('returnHierarchy %s %s %s', hierarchy, projID, showAll)
      self.beSendHierarchy.emit(hierarchy, projDoc)


  @Slot(str, str)
  def returnDoc(self, docID:str) -> None:
    """ Return a document from the database
    Args:
      docID (str): ID of the document to return
      """
    if self.backend is not None:
      self.beSendDoc.emit(self.backend.db.getDoc(docID))


  @Slot(Task, dict)
  def returnTaskReport(self, task:Task, data:dict[str,Any]) -> None:
    """ Handle a rather complicated task request from the GUI and possibly return a report
    Args:
      task (Task): Task to perform
      data (dict): Data required for the task
    """
    if self.backend is None:
      return
    if task is Task.EXTRACTOR_TEST and set(data.keys())=={'fileName','style','recipe','saveFig'}:
      report, image = self.backend.testExtractor(data['fileName'], style={'main':data['recipe']},
                                                 outputStyle=data['style'], saveFig=data['saveFig'])
      self.beSendTaskReport.emit(task, report, image, '')

    elif task is Task.CHECK_DB and set(data.keys())=={'style'}:
      report = self.backend.checkDB(outputStyle=data['style'])
      self.beSendTaskReport.emit(task, report, '', '')

    elif task is Task.SCAN         and set(data.keys())=={'docID'}:
      for _ in range(2):                                                         #scan twice: convert, extract
        self.backend.scanProject(None, data['docID'])
      self.beSendTaskReport.emit(task, 'Scanning finished successfully', '', '')

    elif task is Task.ADD_DOC      and set(data.keys())=={'hierStack','docType','doc'}:
      if data['hierStack']:
        parentID  = data['hierStack'][-1]
        self.backend.cwd = Path(self.backend.db.getDoc(parentID)['branch'][0]['path'])
      else:                                                                                          # project
        self.backend.cwd = Path(self.backend.basePath)
      if '_projectID' in data['doc']:
        del data['doc']['_projectID']
      self.backend.addData(data['docType'], data['doc'], data['hierStack'])
      self.beSendDoc.emit(data['doc'])                                          # send updated doc back to GUI
      if data['docType']=='x0':
        self.beSendTable.emit(self.backend.db.getView('viewDocType/x0'), 'x0')


    elif task is Task.EDIT_DOC      and set(data.keys())=={'doc','newProjID'}:
      # update the path, if the project changed
      if data['newProjID'] and 'branch' in data['doc']:
        parentPath = self.backend.db.getDoc(data['newProjID'][0])['branch'][0]['path']
        if data['doc']['branch'][0]['stack'][0]!=data['newProjID'][0]:                   #only if project changed
          if data['doc']['branch'][0]['path'] is None:
            newPath    = ''
          else:
            oldPath = self.backend.basePath/data['doc']['branch'][0]['path']
            newPath = f'{parentPath}/{oldPath.name}'
          data['doc']['branch'][0] = {'stack':[data['newProjID'][0]], 'path':newPath or None, 'child':9999, 'show':[True,True]}
      # update the doc in the database
      doc = self.backend.db.getDoc(data['doc']['id'])
      if '_projectID' in data['doc']:
        del data['doc']['_projectID']
      doc.update(data['doc'])
      doc = flatten(doc, True)                                                      # type: ignore[assignment]
      self.backend.editData(doc)
      self.beSendDoc.emit(self.backend.db.getDoc(data['doc']['id']))            # send updated doc back to GUI

    elif task is Task.MOVE_LEAVES and set(data.keys())=={'docID','stackOld','stackNew','childOld','childNew'}:
      verbose = False                                                                 # Convenient for testing
      doc      = self.backend.db.getDoc(data['docID'])
      branchOldList= [i for i in doc['branch'] if i['stack']==data['stackOld']]
      if len(branchOldList)!=1:
        logging.error('Cannot move leaves: %s has no branch with stack %s', doc['id'], data['stackOld'], exc_info=True)
        return
      branchOld = branchOldList[0]
      if branchOld['path'] is not None and not branchOld['path'].startswith('http'):
        parentDir = Path(self.backend.db.getDoc(data['stackNew'][-1])['branch'][0]['path'])
        if doc['type'][0][0]=='x':
          dirNameNew= createDirName(doc, data['childNew'], parentDir)# create path name: do not create directory on disk yet
        else:
          dirNameNew= Path(branchOld['path']).name                                              # use old name
        pathNew = f'{parentDir}/{dirNameNew}'
      else:
        pathNew = branchOld['path']
      siblingsNew = self.backend.db.getView('viewHierarchy/viewHierarchy', startKey='/'.join(data['stackNew']))#sorted by docID
      siblingsNew = [i for i in siblingsNew if len(i['key'].split('/'))==len(data['stackNew'])+1]
      childNums   = [f"{i['value'][0]:05d}{i['id']}{idx}" for idx,i in enumerate(siblingsNew)]
      siblingsNew = [x for _, x in sorted(zip(childNums, siblingsNew))]  #sorted by childNum 1st and docID 2nd
      # --- CHANGE ----
      # change new siblings
      if verbose:
        print('\n=============================================\nStep 1: before new siblings')
        print('\n'.join([f'{i["value"][0]} {i["id"]} {i["value"][2]}' for i in siblingsNew]))
      for idx, line in reversed(list(enumerate(siblingsNew))):
        shift = 1 if idx>=data['childNew'] else 0#shift those before the insertion point by 0 and those after by 1
        if line['id']==data['docID'] or line['value'][0]==idx+shift:#ignore this id & those that are correct already
          continue
        if verbose:
          print(f'  {line["id"]}: move: {idx} {shift}')
        self.backend.db.updateBranch(docID=line['id'], branch=line['value'][4], child=idx+shift)
      if verbose:
        siblingsNew = self.backend.db.getView('viewHierarchy/viewHierarchy', startKey='/'.join(data['stackNew']))#sorted by docID
        siblingsNew = [i for i in siblingsNew if len(i['key'].split('/'))==len(data['stackNew'])+1]
        childNums   = [f"{i['value'][0]:05d}{i['id']}{idx}" for idx,i in enumerate(siblingsNew)]
        siblingsNew = [x for _, x in sorted(zip(childNums, siblingsNew))]#sorted by childNum 1st and docID 2nd
        print('Step 2: after new siblings')
        print('\n'.join([f'{i["value"][0]} {i["id"]} {i["value"][2]}' for i in siblingsNew]))
      # change item in question
      if verbose:
        print(f'  manual move {data["childOld"]} -> {data["childNew"]}: {data["docID"]}')
      self.backend.db.updateBranch(docID=data['docID'], branch=-99, stack=data['stackNew'], path=pathNew,
                                   child=data['childNew'], stackOld=data['stackOld']+[data['docID']])
      # change old siblings
      siblingsOld = self.backend.db.getView('viewHierarchy/viewHierarchy', startKey='/'.join(data['stackOld']))#sorted by docID
      siblingsOld = [i for i in siblingsOld if len(i['key'].split('/'))==len(data['stackOld'])+1]
      childNums   = [f"{i['value'][0]:05d}{i['id']}{idx}" for idx,i in enumerate(siblingsOld)]
      siblingsOld = [x for _, x in sorted(zip(childNums, siblingsOld))]#sorted by childNum 1st and docID 2nd
      if verbose:
        print('Step 3: before old siblings')
        print('\n'.join([f'{i["value"][0]} {i["id"]} {i["value"][2]}' for i in siblingsOld]))
      for idx, line in enumerate(siblingsOld):
        if line['value'][0]==idx:                  #ignore id in question and those that are correct already
          continue
        if verbose:
          print(f'  {line["id"]}: move: {idx}')
        self.backend.db.updateBranch(  docID=line['id'], branch=line['value'][4], child=idx)
      if verbose:
        siblingsOld = self.backend.db.getView('viewHierarchy/viewHierarchy', startKey='/'.join(data['stackOld']))#sorted by docID
        siblingsOld = [i for i in siblingsOld if len(i['key'].split('/'))==len(data['stackOld'])+1]
        childNums   = [f"{i['value'][0]:05d}{i['id']}{idx}" for idx,i in enumerate(siblingsOld)]
        siblingsOld = [x for _, x in sorted(zip(childNums, siblingsOld))]#sorted by childNum 1st and docID 2nd
        print('Step 4: end of function')
        print('\n'.join([f'{i["value"][0]} {i["id"]} {i["value"][2]}' for i in siblingsOld]))

    elif task is Task.DROP_EXTERNAL and set(data.keys())=={'docID','files','folders'}:
      commonBase   = os.path.commonpath(data['folders']+[str(i) for i in data['files']])
      doc = self.backend.db.getDoc(data['docID'])
      targetFolder = Path(self.backend.cwd/doc['branch'][0]['path'])
      # create folders and copy files
      for folder in data['folders']:
        (targetFolder/(Path(folder).relative_to(commonBase))).mkdir(parents=True, exist_ok=True)
      for fileStr in data['files']:
        file = Path(fileStr)
        if file.is_file():
          shutil.copy(file, targetFolder/(file.relative_to(commonBase)))
      # scan
      for _ in range(2):                                                       #scan twice: convert, extract
        self.backend.scanProject(None, data['docID'], targetFolder.relative_to(self.backend.basePath))
      self.beSendTaskReport.emit(task, 'Drag-drop operation finished successfully', '', '')

    elif task is Task.DELETE_DOC   and  set(data.keys())=={'docID'}:
      # delete doc: possibly a folder or a project or a measurement
      # delete database and rename folder
      doc = self.backend.db.remove(data['docID'])
      if 'branch' in doc and len(doc['branch'])>0 and 'path' in doc['branch'][0]:
        oldPath = self.backend.basePath/doc['branch'][0]['path']
        newPath = oldPath.parent/f'trash_{oldPath.name}'
        nextIteration = 1
        while newPath.is_dir():
          newPath = oldPath.parent/f'trash_{oldPath.name}_{nextIteration}'
          nextIteration += 1
        if oldPath != newPath:
          oldPath.rename(newPath)
      # go through children, remove from DB
      children = self.backend.db.getView('viewHierarchy/viewHierarchy', startKey=data['docID'])
      for docID in {line['id'] for line in children if line['id']!=data['docID']}:
        self.backend.db.remove(docID)
      # finish it
      if doc['type'][0]=='x0':
        self.beSendTable.emit(self.backend.db.getView('viewDocType/x0'), 'x0')

    elif task is Task.EXPORT_ELN and set(data.keys())=={'fileName','projID','docTypes'}:
      report = exportELN(self.backend, [data['projID']], data['fileName'], data['docTypes'])
      self.beSendTaskReport.emit(task, report, '', '')

    elif task is Task.IMPORT_ELN and set(data.keys())=={'fileName','projID'}:
      report, statistics = importELN(self.backend, data['fileName'], data['projID'])
      self.beSendTaskReport.emit(task, f'{report}\n{json.dumps(statistics,indent=2)}', '', '')

    elif task in (Task.SEND_ELAB, Task.GET_ELAB, Task.SMART_ELAB) and set(data.keys())=={'projGroup'}:
      if 'ERROR' in self.backend.checkDB(minimal=True):
        self.beSendTaskReport.emit(task, 'ERRORs are present in your database. Fix them before uploading', '', '')
      else:
        try:
          sync = Pasta2Elab(self.backend, data['projGroup'])
          if hasattr(sync, 'api') and sync.api.url:                             #if hostname and api-key given
            if task is Task.SEND_ELAB:
              stats = sync.sync('sA')
            elif task is Task.GET_ELAB:
              stats = sync.sync('gA')
            else:
              stats = sync.sync('')
            logging.debug('elabFTW sync stats: %s', stats)
            statsCount = Counter([i[1] for i in stats])
            msg = ', '.join([f'{MERGE_LABELS[k]}: {v}' for k,v in statsCount.items()])
            self.beSendTaskReport.emit(task, f'Success: Synchronized with server. Items per action: {msg}', '', '')
          else:                                                                                  #if not given
            self.beSendTaskReport.emit(task, 'ERROR: Please specify a server address and API-key in the Configuration', '', '')
        except ConnectionError as e:
          logging.error('Connection-Error connecting to elabFTW server')
          self.beSendTaskReport.emit(task, f'ERROR: Connection error to elabFTW server\n{e}', '', '')

    elif task is Task.SEND_TBL_COLUMN and set(data.keys())=={'docType','newList'}:
      self.backend.db.dataHierarchyChangeView(data['docType'], data['newList'])

    elif task is Task.SEND_REPOSITORY and set(data.keys())=={'projID','docTypes','repositories','metadata','uploadZenodo'}:
      tempELN = str(Path(tempfile.gettempdir())/'export.eln')
      res0 = exportELN(self.backend, [data['projID']], tempELN, data['docTypes'])
      print('Export eln',res0)
      repositories = data['repositories']
      if data['uploadZenodo']:                                                                       #Zenodo
        clientZ = ZenodoClient(repositories['zenodo']['url'], repositories['zenodo']['key'])
        metadataZ = clientZ.prepareMetadata(data['metadata'])
        res = clientZ.uploadRepository(metadataZ, tempELN)
      else:                                                                                       #Dataverse
        clientD = DataverseClient(repositories['dataverse']['url'], repositories['dataverse']['key'],
                                repositories['dataverse']['dataverse'])
        metadataD = clientD.prepareMetadata(data['metadata'])
        res = clientD.uploadRepository(metadataD, tempELN)
      msg = 'Successful upload to repository\n'
      # update project with upload details
      if res[0]:
        docProject = self.backend.db.getDoc(data['projID'])
        docProject['.repository_upload'] = f'{datetime.now().strftime("%Y-%m-%d")} {res[1]}'
        docProject['branch'] = docProject['branch'][0] | {'op':'u'}
        self.backend.db.updateDoc(docProject, data['projID'])
        msg += 'Saved information to project'
      else:
        msg += 'Error while writing project information to database'
      self.beSendTaskReport.emit(task, msg, '', '')

    elif task is Task.EXTRACTOR_RERUN and set(data.keys())=={'docIDs','recipe'}:
      for docID in data['docIDs']:
        doc = self.backend.db.getDoc(docID)
        if data['recipe']:
          doc['type'] = data['recipe'].split('/')
        #any path is good since the file is the same everywhere; data-changed by reference
        if doc['branch'][0]['path'] is not None:
          oldDocType = doc['type']
          doc['type'] = ['']
          if doc['branch'][0]['path'].startswith('http'):
            path = Path(doc['branch'][0]['path'])
          else:
            path = self.backend.basePath/doc['branch'][0]['path']
          self.backend.useExtractors(path, doc.get('shasum',''), doc)
          if doc['type'][0] == oldDocType[0]:
            del doc['branch']                                                                  #don't update
            self.backend.db.updateDoc(doc, docID)
          else:
            self.backend.db.remove( docID )
            del doc['id']
            doc['name'] = doc['branch'][0]['path']
            self.backend.addData('/'.join(doc['type']), doc, doc['branch'][0]['stack'])
      self.beSendTaskReport.emit(task, 'Extractors re-ran successfully', '', '')

    elif task is Task.OPEN_EXTERNAL and set(data.keys())=={'docID'}:
      doc   = self.backend.db.getDoc(data['docID'])
      if doc['branch'][0]['path'] is None:
        QMessageBox.critical(None, 'ERROR', 'Cannot open file that is only in the database')
      else:
        path  = Path(self.backend.basePath)/doc['branch'][0]['path']
        self.beSendTaskReport.emit(task, '', '', str(path))

    elif task is Task.SET_GUI    and set(data.keys())=={'docID','gui'}:
      self.backend.db.setGUI(data['docID'], data['gui'])

    elif task is Task.HIDE_SHOW    and set(data.keys())=={'docID'}:
      self.backend.db.hideShow(data['docID'])

    else:
      logging.error('Got task, which I do not understand %s %s', task, data.keys(), exc_info=True)


  @Slot(list)
  def executeSQL(self, tasks:list[dict[str,Any]]) -> None:
    """ Execute SQL commands in the backend database: fast change
    Args:
      tasks (list): List of tasks to execute, each task is a dict with keys:
                    - type: 'one', 'many', 'df', or 'get_df'
                    - cmd: SQL command to execute
                    - list: List of parameters for the command (optional)
    """
    if self.backend is None:
      return
    for task in tasks:
      logging.debug('Executing SQL task %s', task)
      if task['type']=='one':
        self.backend.db.cursor.execute(task['cmd'], task.get('list',()))
      elif task['type']=='many':
        self.backend.db.cursor.executemany(task['cmd'], task['list'])
      elif task['type']=='df':
        task['df'].to_sql(task['table'], self.backend.db.connection, if_exists='append', index=False, dtype='str')
      elif task['type']=='get_df':
        df = pd.read_sql_query(task['cmd'], self.backend.db.connection).fillna('')
        self.beSendSQL.emit(task['cmd'], df)
      else:
        print('**ERROR unknown task command', task)
    self.backend.db.connection.commit()


  def exit(self) -> None:
    """ Exit the worker thread """
    if self.backend is not None:
      self.deleteLater()


class BackendThread(QThread):
  """
  Thread that manages the backend worker
  """
  def __init__(self, parent:QObject|None=None) -> None:
    """Initialize the backend thread with a parent QObject
    Args:
      parent (QObject): Parent QObject for the thread
    """
    super().__init__(parent)
    self.worker = BackendWorker()
    self.worker.moveToThread(self)
    self.finished.connect(self.worker.exit)

  def run(self) -> None:
    """
    Run the thread event loop
    """
    self.exec()
